


#include "disk.h"
#include "graphics.h"
#include "mem.h"
#include "everdrive.h"
#include "sega.h"

#define CMD0  0x40    // software reset
#define CMD1  0x41    // brings card out of idle state
#define CMD8  0x48    // Reserved
#define CMD12 0x4C    // stop transmission on multiple block read
#define CMD17 0x51    // read single block
#define CMD18 0x52    // read multiple block
#define CMD58 0x7A    // reads the OCR register
#define CMD55 0x77
#define CMD41 0x69
#define CMD23 0x57
#define CMD24 0x58    // writes a single block
#define CMD25 0x59    // writes a single block
#define	ACMD23 (23 | 0x80)
#define SD_V2 2
#define SD_HC 1

u8 *card_type;

u8 sd_card_type;
u8 msd_card_type;

void diskCloseRW();
u8 diskOpenRead(u32 saddr);

u8 mmcCmd(u8 cmd, u32 arg);
u8 mmcCmdCrc(u8 cmd, u32 arg, u8 crc);
u8 diskInitStd();
u8 diskInitHC();

u8 mmcCmd(u8 cmd, u32 arg) {

    return mmcCmdCrc(cmd, arg, 0x95);
}

u8 mmcCmdCrc(u8 cmd, u32 arg, u8 crc) {

    u16 i = 0;
    u8 resp;
    memSpiBusy();
    memSpiSSOff();
    memSpiSSOn();
    mem_spi(0xff);
    mem_spi(cmd);
    mem_spi(arg >> 24);
    mem_spi(arg >> 16);
    mem_spi(arg >> 8);
    mem_spi(arg);
    mem_spi(crc);
    mem_spi(0xff);
    resp = mem_spi(0xff);

    while (resp == 0xff) {
        resp = mem_spi(0xff);
        if (i++ == 2048)break;
    }
    /*
        if (i > 1) {
            gConsDrawString("cmd_len: ");
            gAppendNum(i);
            while (joy == 0);
            while (joy != 0);
        }*/

    memSpiSSOff();
    return resp;
}

u8 diskInit() {

    u8 resp;

    //consDrawStr("HC");
    SPI_CFG_PORT &= (1 << _SPI_DISK);
    resp = diskInitHC();


    if (resp == DISK_ERR_INIT)return resp;
    if (resp == 0)return 0;
    //consDrawStr("ST");
    //resp = diskInitStd();

    return resp;

}

u8 diskInitStd() {

    u16 i;
    *card_type = 0;
    memSpiSSOff();
    memSpiSetSpeed(SPI_SPEED_INIT);
    SPI16_OFF;
    for (i = 0; i < 40; i++) {
        mem_spi(0xff);
    }

    if (mmcCmd(CMD0, 0) != 1) {
        return DISK_ERR_INIT + 0;
    }

    i = 0;

    while (mmcCmd(CMD1, 0) != 0) {
        if (i++ == 65535) {
            return DISK_ERR_INIT + 1;
        }
    }

    memSpiSSOn();
    mem_spi(0xff);

    memSpiSetSpeed(SPI_SPEED_50);
    memSpiSSOff();

    return 0;
}

u8 diskInitHC() {

    u16 i;
    volatile u8 resp = 0;
    u8 cmd;
    u32 wait_len = 32768;



    *card_type = 0;



    memSpiSSOff();
    SPI16_OFF;
    memSpiSetSpeed(SPI_SPEED_INIT);


    // for (i = 0; i < 2; i++) {
    i = 0;
    while (resp != 1 && i++ < 2) {
        resp = mmcCmdCrc(CMD0, 0, 0x95);
    }

    // if (resp == 1)break;
    // }
    if (resp != 1) return DISK_ERR_INIT + 0;


    //mmcCmdCrc(0x7b, 0, 0x91);
    resp = mmcCmdCrc(CMD8, 0x1aa, 0x87);
    for (i = 0; i < 5; i++)mem_spi(0xff);
    if (resp == 0xff)return DISK_ERR_INIT + 1;
    if (resp != 5)*card_type |= SD_V2;

    if (*card_type == 2) {

        for (i = 0; i < wait_len; i++) {


            resp = mmcCmd(CMD55, 0xffff);
            if (resp == 0xff)return DISK_ERR_INIT + 2;
            if (resp != 1)continue;

            resp = mmcCmd(CMD41, 0x40300000);
            if (resp == 0xff)return DISK_ERR_INIT + 3;
            if (resp != 0)continue;

            break;
        }
        if (i == wait_len)return DISK_ERR_INIT + 4;

        resp = mmcCmd(CMD58, 0);
        if (resp == 0xff)return DISK_ERR_INIT + 5;
        memSpiSSOn();
        resp = mem_spi(0xff);
        for (i = 0; i < 3; i++)mem_spi(0xff);
        if ((resp & 0x40))*card_type |= 1;
    } else {


        i = 0;

        resp = mmcCmd(CMD55, 0);
        if (resp == 0xff)return DISK_ERR_INIT + 6;
        resp = mmcCmd(CMD41, 0);
        if (resp == 0xff)return DISK_ERR_INIT + 7;
        cmd = resp;

        for (i = 0; i < wait_len; i++) {
            if (resp < 1) {

                resp = mmcCmd(CMD55, 0);
                if (resp == 0xff)return DISK_ERR_INIT + 8;
                if (resp != 1)continue;

                resp = mmcCmd(CMD41, 0);
                if (resp == 0xff)return DISK_ERR_INIT + 9;
                if (resp != 0)continue;

            } else {

                resp = mmcCmd(CMD1, 0);
                if (resp != 0)continue;
            }

            break;

        }

        if (i == wait_len)return DISK_ERR_INIT + 10;

    }

    memSpiSetSpeed(SPI_SPEED_50);

    return 0;
}

u8 diskReadSector(u32 saddr, void *dat) {

    u8 resp;

    //gConsDrawString("addr: ");
    //gAppendNum(saddr);

    if (!(*card_type & 1))saddr *= 512;

    resp = mmcCmd(CMD17, saddr);
    if (resp != 0)return DISK_ERR_RD1;
    memSpiSSOn();



    resp = memSpiRead(dat, 1);

    memSpiSSOff();
    return resp;

}

u8 diskOpenRead(u32 saddr) {

    u8 resp;
    if (!(*card_type & 1))saddr *= 512;



    resp = mmcCmd(CMD18, saddr);
    if (resp != 0)return DISK_ERR_RD2;
    memSpiSSOn();

    return 0;
}

void diskCloseRW() {

    memSpiSSOff();
    mmcCmd(CMD12, 0);
}

u8 diskRead(u32 saddr, void *buff, u16 slen) {


    u8 resp;
    resp = diskOpenRead(saddr);
    if (resp)return resp;


    resp = memSpiRead(buff, slen);

    diskCloseRW();

    return resp;
}

u8 diskWriteSector(u32 saddr, u8 *buff) {

    //return 0;
    u16 i;
    u8 resp = 0;
    if (!(*card_type & 1))saddr *= 512;

    resp = mmcCmd(CMD24, saddr);
    if (resp != 0)return DISK_ERR_WR1;
    memSpiSSOn();

    mem_spi(0xff);
    mem_spi(0xff);
    mem_spi(0xfe);

    memSpiWrite(buff);

    mem_spi(0xff);
    mem_spi(0xff);
    resp = mem_spi(0xff);
    //return 0;

    if ((resp & 0x1f) != 0x05) {
        memSpiSSOff();
        return DISK_ERR_WR2;
    }


    for (i = 0; i < 65535; i++) {
        if (mem_spi(0xff) == 0xff)break;
    }
    memSpiSSOff();
    if (i == 65535)return DISK_ERR_WR3;

    return 0;
}

void diskSelect(u8 disk) {

    if (disk == DISK_MSD) {
        SPI_CFG_PORT |= 1 << _SPI_DISK;
        card_type = &msd_card_type;
    } else {
        SPI_CFG_PORT &= ~(1 << _SPI_DISK);
        card_type = &sd_card_type;
    }
}

u16 diskGetCurrent() {

    u16 disk = SPI_CFG_PORT;
    disk &= (1 << _SPI_DISK);
    if (disk != 0) {
        disk = DISK_MSD;
    } else {
        disk = DISK_SD;
    }
    return disk;
}

u8 diskOpenWrite(u32 saddr) {

    u8 resp;
    if (!(*card_type & 1))saddr *= 512;

    resp = mmcCmd(CMD25, saddr);
    if (resp != 0)return DISK_ERR_WR4;
    memSpiSSOn();

    return 0;
}

u8 diskWrite(u32 saddr, u8 *buff, u16 slen) {

    u8 resp;
    u16 i;

    // resp = mmcCmd(ACMD23, slen);
    //gConsDrawString("resp: ");
    //gAppendNum(resp);
    //if (resp != 0)return DISK_ERR_WR5;

    resp = diskOpenWrite(saddr);


    while (slen--) {

        mem_spi(0xff);
        mem_spi(0xff);
        mem_spi(0xfc);


        memSpiWrite(buff);

        mem_spi(0xff);
        mem_spi(0xff);
        resp = mem_spi(0xff);
        //resp = mem_spi(0xff);

        buff += 512;



        if ((resp & 0x1f) != 0x05) {
            memSpiSSOff();
            return DISK_ERR_WR2;
        }


        for (i = 0; i < 65535; i++) {

            if (mem_spi(0xff) == 0xff)break;
            if (mem_spi(0xff) == 0xff)break;
        }


        /*
                if (i > 2) {
                    gConsDrawString("pop: ");
                    gAppendNum(i);
                }*/

        if (i == 65535) {
            memSpiSSOff();
            return DISK_ERR_WR3;
        }

    }

    mem_spi(0xfd);
    mem_spi(0xff);

    for (i = 0; i < 65535; i++) {
        if ((mem_spi(0xff) & 1) != 0)break;
    }
    memSpiSSOff();
    if (i == 65535) return DISK_ERR_WR5;
    //diskCloseRW();

    return 0;
}

/*

{
    //multiple sector write
    SendCMD(ACMD23, lenght);
    spiSendByte(CMD25 | 0x40);
    spiSendByte(start >> 24);
    spiSendByte(start >> 16);
    spiSendByte(start >> 8);
    spiSendByte(start);
    spiSendByte(0xff);

    for (i = 0;; i++) {
        resp = spiSendByte();
        if (!(resp & 0x80)) {
            if (resp) BSOD("SD card IO", "CMD25, error: 0x%x", resp);
            break;
        }
        if (i >= SD_NCR) BSOD("SD card IO", "CMD25, card not responding.");
    }
    spiSendByte(); //skip dummy byte
    while (lenght--) {
        spiSendByte(DATA_TOKEN_FC);
        DMA_WriteSector(buffer);
        buffer += 512;
        spiSendByte();
        spiSendByte();
        for (i = 0;; i++) {
            resp = spiSendByte();
            if (resp != 0xff) {
                if ((resp & 0x1f) == DATA_RESP_TOKEN) break;
                else BSOD("SD card IO", "CMD25, error token: 0x%x", resp);
            }
            if (i >= SD_NCR) BSOD("SD card IO", "CMD25, missing resp_token");
        }
        timeout = 250 ms;
        while (1) {
            token = spiSendByte();
            if (token & 0x1) break;
            if (!timeout) BSOD("SD card IO", "CMD25, write timeout!");
        }
    }
    spiSendByte(STOP_TR_TOKEN);
    spiSendByte();
    timeout = 250 ms;
    while (1) {
        token = spiSendByte();
        if (token & 0x1) break;
        if (!timeout) BSOD("SD card IO", "CMD25, write timeout!.");
    }
}
spiSendByte(CMD13 | 0x40);
spiSendByte(0);
spiSendByte(0);
spiSendByte(0);
spiSendByte(0);
spiSendByte(0xff);

for (i = 0;; i++) {
    resp = spiSendByte();
    if (!(resp & 0x80)) {
        resp <<= 8;
        resp |= spiSendByte();
        if (resp) BSOD("SD card IO", "CMD13, error: 0x%x", resp);
        break;
    }
    if (i >= SD_NCR) BSOD("SD card IO", "CMD13, card not responding.");
}
spiSendByte()*/
