

#include "fat.h"
#include "disk.h"
#include "mem.h"
#include "string.h"
#include "sega.h"
#include "graphics.h"
#include "everdrive.h"
#include "os.h"
#include "gui.h"

//#include "tools_.h"




Dir *dir;

FatCache *fat_cache;
Fat *current_fat;
File *current_file;

void fatMalloc() {

    dir = memAllocate(sizeof (Dir), 4);
    fat_cache = (FatCache *) memAllocate(sizeof (FatCache), 512);
    current_fat = memAllocate(sizeof (Fat), 4);
}

u8 fatSlen(u8 *str);
u8 fatLoadLfnPart(u8 *name_ptr, u8 *lfn_buff);
u8 fatCacheLoadTable(u32 sector);
u8 fatCacheLoadData(u32 sector);

u8 fatGetNextCluster(u32 *cluster);
u8 fatGetTableRecord(u32 cluster, u32 *val);
u8 fatSetTableRecord(u32 cluster, u32 next_cluster);
u8 fatGetFreeRecord(FatFullRecord *dir_rec, u8 len);
void fatMakeLfnBlock(u8 *name, u8 *lfn_block, u8 block_idx, u8 crc);
u32 fatSectorToCluster(u32 sector);
void fatStrHiCase(u8 *str, u8 len);
u8 fatClearClusters(u32 cluster, u8 len);
void fatSortRecords();
u8 fatReadCluster(void *dst);
u8 fatOccupyClusters(u32 *base_cluster, u16 len);

u32 fatClusterToSector(u32 cluster);
u8 fatCacheSaveData();
u8 fatCacheSaveTable();
u8 fatResizeFile(FatFullRecord *rec, u32 req_sec_size);
u16 fatGetRecordName(FatRecord *rec, u8 *name, u16 max_len);
u8 fatCreateRecord(u8 *name, u8 is_dir, u8 check_exist);
u8 fatFindRecord(u8 *name, FatFullRecord *rec, u8 is_dir);

u16 fatStrEq(u8 *str1, u8 *str2) {

    u8 s1;
    u8 s2;

    for (;;) {
        s1 = *str1++;
        s2 = *str2++;
        if (s1 >= 'a' && s1 <= 'z')s1 -= 0x20;
        if (s2 >= 'a' && s2 <= 'z')s2 -= 0x20;

        if (s1 != s2) return 0;

        if (*str1 == 0 && *str2 == 0)return 1;
    }
}

u16 fatStrEqL(u8 *str1, u8 *str2, u8 len) {

    u8 s1;
    u8 s2;
    while (len--) {

        s1 = *str1++;
        s2 = *str2++;
        if (s1 >= 'a' && s1 <= 'z')s1 -= 0x20;
        if (s2 >= 'a' && s2 <= 'z')s2 -= 0x20;

        if (s1 != s2) return 0;
    }

    return 1;
}

u8 fatSlen(u8 *str) {

    u8 len = 0;
    while (*str++)len++;
    return len;
}

void fatStrHiCase(u8 *str, u8 len) {

    if (len) {
        while (len--) {
            if (*str >= 'a' && *str <= 'z')*str -= 0x20;
            str++;
        }
    } else {
        while (*str != 0) {
            if (*str >= 'a' && *str <= 'z')*str -= 0x20;
            str++;
        }
    }

}

u8 fatCreateRecIfNotExist(u8 *name, u8 is_dir) {


    FatFullRecord rec_tmp;
    u8 resp;

    resp = fatFindRecord(name, &rec_tmp, is_dir);

    if (resp == FAT_ERR_NOT_EXIST) {
        resp = fatCreateRecord(name, is_dir, 0);
        if (resp)return resp;
    } else
        if (resp)return resp;

    return 0;
}

u8 fatInit() {

    // u16 i;

    u8 resp;


    u32 reserved_sectors;
    //if (disk > 1)disk = 1;

    //fat[disk]->fat_ok = 0;

    //diskSelect(disk);
    resp = diskInit();
    if (resp)return resp;


    //current_fat = fat[disk];
    //current_cache = cache[disk];
    fat_cache->data_sec_idx = 0xffffffff;
    fat_cache->table_sec_idx = 0xffffffff;
    fat_cache->table_changed = 0;
    //current_cache->table_frame = current_cache->table_buff;
    //current_cache->data_frame = current_cache->data_buff;
    current_fat->pbr_entry = 0;


    resp = fatCacheLoadData(current_fat->pbr_entry);
    if (resp)return resp;


    if (!fatStrEqL("FAT16", &fat_cache->data_sector[0x36], 5) && !fatStrEqL("FAT32", &fat_cache->data_sector[0x52], 5)) {

        current_fat->pbr_entry = (u32) fat_cache->data_sector[0x1c6] | (u32) fat_cache->data_sector[0x1c7] << 8 | (u32) fat_cache->data_sector[0x1c8] << 16 | (u32) fat_cache->data_sector[0x1c9] << 24;
        resp = fatCacheLoadData(current_fat->pbr_entry);
        if (resp)return resp;
        if (!fatStrEqL("FAT16", &fat_cache->data_sector[0x36], 5) && !fatStrEqL("FAT32", &fat_cache->data_sector[0x52], 5)) {
            return FAT_ERR_INIT;
        }
    }

    if (fatStrEqL("FAT16", &fat_cache->data_sector[0x36], 5)) {
        current_fat->type = FAT_TYPE_16;
        current_fat->sectors_per_fat = (u32) fat_cache->data_sector[0x16] | (u32) fat_cache->data_sector[0x17] << 8;
    } else {
        current_fat->type = FAT_TYPE_32;
        current_fat->sectors_per_fat = (u32) fat_cache->data_sector[0x24] | (u32) fat_cache->data_sector[0x25] << 8 | (u32) fat_cache->data_sector[0x26] << 16 | (u32) fat_cache->data_sector[0x27] << 24;
    }
    reserved_sectors = (u32) fat_cache->data_sector[0x0e] | (u32) fat_cache->data_sector[0x0f] << 8;

    current_fat->cluster_size = fat_cache->data_sector[0x0d];
    current_fat->root_entry = current_fat->sectors_per_fat * 2 + current_fat->pbr_entry + reserved_sectors;
    current_fat->fat_entry = current_fat->pbr_entry + reserved_sectors;
    current_fat->data_entry = current_fat->root_entry;
    if (current_fat->type == FAT_TYPE_16) {
        current_fat->data_entry += 16384 / 512;
    }
    current_fat->cluster_byte_size = current_fat->cluster_size * 512;


    //resp = fatInitRoot();
    //if (resp)return resp;

    //fat[disk]->fat_ok = 1;

    return 0;
}

void fatOpenDir(u32 cluster) {


    if (cluster == 0) {
        dir->entry_sector = current_fat->root_entry;
        dir->is_root = 1;
        if (current_fat->type == FAT_TYPE_32)cluster = 2;
    } else {
        dir->entry_sector = fatClusterToSector(cluster);
        dir->is_root = 0;
    }

    dir->entry_cluster = cluster;
    dir->current_cluster = cluster;
    dir->in_cluster_ptr = 0;

}

u8 fatLoadDir(u32 cluster) {

    u8 resp;

    fatOpenDir(cluster);
    for (dir->size = 0; dir->size < FAT_MAX_DIR_SIZE; dir->size++) {

        resp = fatReadDir(&dir->rec[dir->size]);
        // gAppendString(dir->rec[dir->size].name_piece);
        if (resp)return resp;
        if (dir->rec[dir->size].hdr_sector == 0)break;

    }


    fatSortRecords();


    return 0;
}

void fatSortRecords() {


    FatRecord *rec_buff;
    u16 i;
    u8 flag0;
    u8 flag1;
    //u32 v1;
    //u32 v2;


    // VDP_drawText(APLAN, "1", 0, 2, cy++);
    for (i = 0; i < dir->size; i++) {
        dir->s_records[i] = &dir->rec[i];
        *(u32 *) dir->s_records[i]->name_begin |= 0x20202020;
    }


    // v1 = *(u32 *) dir->s_records[k]->name_begin;
    //v2 = *(u32 *) dir->s_records[k + x]->name_begin;

    s16 increment = dir->size;
    s16 j;
    u32 temp;
    while (increment > 0) {
        for (i = increment; i < dir->size; i++) {
            j = i;
            temp = *(u32 *) dir->s_records[i]->name_begin;
            rec_buff = dir->s_records[i];

            while (j >= increment && *(u32 *) dir->s_records[j - increment]->name_begin > temp) {
                dir->s_records[j] = dir->s_records[j - increment];
                j = j - increment;
            }
            dir->s_records[j] = rec_buff;
        }
        if (increment == 2) {
            increment = 1;
        } else {
            increment = increment * 5 / 11;
        }
    }




    for (i = 1; i < dir->size;) {


        flag0 = dir->s_records[i - 1]->is_dir;
        flag1 = dir->s_records[i]->is_dir;

        if (flag1 && !flag0) {
            rec_buff = dir->s_records[i];
            dir->s_records[i] = dir->s_records[i - 1];
            dir->s_records[i - 1] = rec_buff;
            if (i > 1)i--;
        } else {
            i++;
        }
    }
    //VDP_drawText(APLAN, "4", 0, 2, cy++);
    /*
        int ctr = 0;
        if (dir->size > 1) {
            for (i = 0; i < dir->size - 1; i++) {
                if (dir->s_records[i]->is_dir != dir->s_records[i + 1]->is_dir)continue;

                v1 = *(u32 *) dir->s_records[i]->name_begin;
                v2 = *(u32 *) dir->s_records[i + 1]->name_begin;

                if (v1 > v2) {
                    ctr++;

                    guiShowMessage("mes", "sort error");
                    JOY_waitAct();
                    return;

                }

            }

        }*/
    /*
        guiShowMessage("mes", "OK!");
        JOY_waitAct();*/

}

u8 fatReadDir(FatRecord *rec) {


    u16 in_sector_ptr = 0;
    u32 current_sector = 0;
    u16 i = 0;
    u8 *ptr = 0;
    u8 resp;
    u32 base_sector; // = dir->entry_sector == dir->is_root ? fat->root_entry : fatClusterToSector(dir->current_cluster);
    u8 apple_file = 0;

    // rec->lfn_len = 0;

    rec->name_begin[0] = 0;
    rec->hdr_idx = 0;
    rec->hdr_sector = 0;

    if (current_fat->type == FAT_TYPE_16 && dir->is_root) {
        base_sector = current_fat->root_entry;
    } else {
        base_sector = fatClusterToSector(dir->current_cluster);
    }


    for (;;) {

        current_sector = base_sector + dir->in_cluster_ptr / 512;

        if (dir->is_root && current_fat->type == FAT_TYPE_16 && dir->in_cluster_ptr >= 16384) {
            return 0;
        }

        if (dir->in_cluster_ptr >= current_fat->cluster_byte_size && (!dir->is_root || current_fat->type == FAT_TYPE_32)) {

            resp = fatGetNextCluster(&dir->current_cluster);
            if (resp)return resp;
            if (dir->current_cluster == 0)return 0;

            base_sector = fatClusterToSector(dir->current_cluster);
            current_sector = base_sector; // + dir->in_cluster_ptr / 512;
            dir->in_cluster_ptr = 0;
        }

        in_sector_ptr = dir->in_cluster_ptr & 511;
        resp = fatCacheLoadData(current_sector);
        if (resp)return resp;

        if (fat_cache->data_sector[in_sector_ptr] == 0) return 0;

        ptr = &fat_cache->data_sector[in_sector_ptr];

        if (*ptr == 0xe5 || *ptr == 0x2e) {
            dir->in_cluster_ptr += 32;
            apple_file = 0;
            rec->name_begin[0] = 0;
            continue;
        }

        if (ptr[0x0b] == 0x0f) {

            dir->in_cluster_ptr += 32;
            apple_file = ptr[1] == '.' && (ptr[0] & 0xf) == 1 ? 1 : 0;

            if ((ptr[0] & 0xf) == 1) {
                rec->name_begin[0] = ptr[1];
                rec->name_begin[1] = ptr[3];
                rec->name_begin[2] = ptr[5];
                rec->name_begin[3] = ptr[7];
            } else {
                rec->name_begin[0] = 0;
            }

            continue;
        }

        if (apple_file) {
            dir->in_cluster_ptr += 32;
            apple_file = 0;
            continue;
        }

        ptr = &fat_cache->data_sector[in_sector_ptr];
        if (rec->name_begin[0] == 0) {
            for (i = 0; i < 4; i++)rec->name_begin[i] = *ptr++;
        }


        ptr = &fat_cache->data_sector[in_sector_ptr];
        rec->is_dir = (ptr[0x0b] & 0x10) == 0x10 ? 1 : 0;


        rec->hdr_idx = (dir->in_cluster_ptr & 511) / 32;
        rec->hdr_sector = current_sector;

        dir->in_cluster_ptr += 32;
        return 0;


    }


    return 1;
}

u8 fatCacheLoadData(u32 sector) {

    u8 resp = 0;

    if (fat_cache->data_sec_idx == sector)return 0;

    resp = diskRead(sector, fat_cache->data_sector, 1);

    fat_cache->data_sec_idx = sector;
    return resp;

}

u8 fatCacheSaveData() {

    return diskWrite(fat_cache->data_sec_idx, fat_cache->data_sector, 1);
}

u8 fatCacheLoadTable(u32 sector) {

    u8 resp = 0;
    //sector += fat->fat_entry;
    if (fat_cache->table_sec_idx == sector)return 0;

    if (fat_cache->table_changed) {
        resp = fatCacheSaveTable();
        if (resp)return resp;
    }

    resp = diskRead(sector + current_fat->fat_entry, fat_cache->table_sector, 1);

    fat_cache->table_sec_idx = sector;

    return resp;
}

u8 fatCacheSaveTable() {

    u8 resp;

    if (!fat_cache->table_changed)return 0;

    fat_cache->table_changed = 0;

    resp = diskWriteSector(fat_cache->table_sec_idx + current_fat->fat_entry, fat_cache->table_sector);
    if (resp)return resp;
    resp = diskWriteSector(fat_cache->table_sec_idx + current_fat->fat_entry + current_fat->sectors_per_fat, fat_cache->table_sector);
    return resp;
}

u32 fatClusterToSector(u32 cluster) {

    return (cluster - 2) * current_fat->cluster_size + current_fat->data_entry;
}

u32 fatSectorToCluster(u32 sector) {

    return (sector - current_fat->data_entry) / current_fat->cluster_size + 2;
}

u8 fatGetNextCluster(u32 *cluster) {

    u8 resp;


    resp = fatGetTableRecord(*cluster, cluster);
    if (resp)return resp;

    if (current_fat->type == FAT_TYPE_16) {
        if (*cluster == 0xffff) {
            *cluster = 0;
        }
    } else {
        if (*cluster == 0x0FFFFFFF) {
            *cluster = 0;
        }
    }
    return 0;

}

u8 fatOpenFileByeName(u8 *name, u32 wr_sectors) {


    u8 resp;
    FatFullRecord frec;


    resp = fatFindRecord(name, &frec, 0);
    if (resp)return resp;
    if (frec.is_dir)return FAT_ERR_NOT_FILE;
    resp = fatOpenFile(&frec, wr_sectors);
    if (resp)return resp;

    return 0;
}

u8 fatFindRecord(u8 *name, FatFullRecord *frec, u8 is_dir) {

    FatRecord rec;
    u8 resp;
    u8 *sub_name_ptr;
    u8 sub_name[256];
    u8 name_len;


    // u16 i;
    //u16 sub_name_len;


    frec->entry_cluster = 0;

    for (;;) {

        if (*name == '/')name++;
        sub_name_ptr = sub_name;
        name_len = 0;
        while (*name != 0 && *name != '/') {
            *sub_name_ptr++ = *name++;
            name_len++;
        }

        *sub_name_ptr = 0;

        fatOpenDir(frec->entry_cluster);


        for (;;) {


            resp = fatReadDir(&rec);


            if (resp)return resp;
            if (rec.hdr_sector == 0) {

                return FAT_ERR_NOT_EXIST;
            }


            if (rec.name_begin[3] != 0 && rec.name_begin[3] != 0x20) {
                if (!fatStrEqL(sub_name, rec.name_begin, 4))continue;
            }



            //fatGetRecordName(rec, rec_full_name, 0);
            fatGetFullRecord(&rec, frec);



            if (fatStrEq(sub_name, frec->name)) {

                if (*name != 0)break;
                if (*name == 0 && ((is_dir && rec.is_dir) || (!is_dir && !rec.is_dir)))break;
            }
        }


        if (*name == 0)return 0;

    }

    return 0;

}

//FatRecordHdr hdr;

/*
void fatGetNameExt(u8 *filename, u8 *name, u8 *ext){

}
 */


void fatSetClustVal(FatRecordHdr *hdr, u32 clust_val) {

    ((u8 *) & hdr->cluster_lo)[0] = (u32) clust_val & 0xff;
    ((u8 *) & hdr->cluster_lo)[1] = (u32) clust_val >> 8;
    //hdr.cluster_lo = cluster & 0xffff;
    if (current_fat->type == FAT_TYPE_32) {
        ((u8 *) & hdr->cluster_hi)[0] = (u32) clust_val >> 16;
        ((u8 *) & hdr->cluster_hi)[1] = (u32) clust_val >> 24;
    }

}

/*
static const u8 rawData[32] = {
    0x2E, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x10, 0x00, 0x34, 0xD3, 0xB9,
    0x53, 0x40, 0x53, 0x40, 0x00, 0x00, 0xD4, 0xB9, 0x53, 0x40, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00,
};*/

u8 fatCreateRecord(u8 *name, u8 is_dir, u8 check_exist) {

    //u8 cy = 10;
    u16 i;
    FatRecordHdr hdr;
    FatFullRecord rec;
    u8 patch[256];
    u8 *file_name;
    u8 *ptr;
    u8 lfn_blocks = 0;
    u8 resp;
    u8 sum = 0;
    u8 nam_len;
    u8 ext_len;
    u32 cluster;
    u8 lfn_block[32];
    u32 sub_dir_cluster = 0;
    u16 name_len = fatSlen(name);

    fat_cache->data_sec_idx = ~0;

    if (*name != '/')return FAT_ERR_NAME;
    if (name_len < 2)return FAT_ERR_NAME;
    i = name_len - 1;
    if (name[i] == '.' || name[i] == ' ')return FAT_ERR_NAME;


    if (check_exist) {
        resp = fatFindRecord(name, &rec, is_dir);
        if (resp != FAT_ERR_NOT_EXIST) {
            if (resp)return resp;
            return FAT_ERR_EXIST;
        }
    }



    memfill(&hdr, 0, 32);


    ptr = &name[name_len];

    i = 0;
    while (*ptr != '/')ptr--;

    i = 0;
    while (name != ptr)patch[i++] = *name++;
    patch[i] = 0;

    ptr++;
    file_name = ptr;
    //i = 0;
    //while (*ptr != 0)file_name[i++] = *ptr++;
    //file_name[i] = 0;

    name_len = fatSlen(file_name);

    if (name_len < 1)return FAT_ERR_NAME;


    i = name_len;
    while (i != 0 && file_name[i] != '.')i--;
    if (i == 0)i = name_len;
    nam_len = i;
    ext_len = name_len;
    if (ext_len == nam_len) {
        ext_len = 0;
    } else {
        ext_len -= (nam_len + 1);
    }



    if (nam_len > 8 || ext_len > 3) {

        lfn_blocks = name_len;
        if (lfn_blocks % 13 != 0)lfn_blocks += 13;
        lfn_blocks /= 13;
        for (i = 0; i < 8; i++)hdr.name[i] = file_name[i];
    } else {
        for (i = 0; i < 8 && file_name[i] != 0 && file_name[i] != '.'; i++)hdr.name[i] = file_name[i];
        for (; i < 8; i++)hdr.name[i] = ' ';
    }
    //VDP_drawText(APLAN, file_name, 0, 2, cy++);

    ptr = 0;
    if (ext_len != 0)ptr = &file_name[nam_len + 1];


    if (ptr != 0) {
        hdr.ext[0] = *ptr == 0 ? 0x20 : *ptr++;
        hdr.ext[1] = *ptr == 0 ? 0x20 : *ptr++;
        hdr.ext[2] = *ptr == 0 ? 0x20 : *ptr++;
    } else {
        for (i = 0; i < 3; i++)hdr.ext[i] = ' ';
    }
    fatStrHiCase(hdr.name, 8);
    fatStrHiCase(hdr.ext, 3);
    hdr.attrib = is_dir ? 0x10 : 0x20;




    if (patch[0] != 0) {
        resp = fatFindRecord(patch, &rec, 1);
        sub_dir_cluster = rec.entry_cluster;
        //if (resp == FAT_ERR_NOT_EXIST)VDP_drawText(APLAN, "cant find patch", 0, 2, cy++);
        if (resp)return resp;
    } else {
        rec.entry_cluster = 0;
    }


    resp = fatGetFreeRecord(&rec, lfn_blocks + 1);
    if (resp)return resp;

    //rec.hdr_sector = fat->root_entry;
    //rec.hdr_idx


    resp = fatCacheLoadData(rec.hdr_sector);
    if (resp)return resp;

    if (lfn_blocks != 0) {

        hdr.name[7] = '1';
        hdr.name[6] = '~';
        hdr.name[5] = (rec.hdr_idx + rec.hdr_sector / 32) & 15;
        hdr.name[4] = ((rec.hdr_idx + rec.hdr_sector / 32) >> 4) & 15;
        hdr.name[5] += hdr.name[5] <= 10 ? 48 : 55;
        hdr.name[4] += hdr.name[4] <= 10 ? 48 : 55;
        for (sum = i = 0; i < 11; i++) {
            if (i < 8) {
                sum = (((sum & 1) << 7) | ((sum & 0xfe) >> 1)) + hdr.name[i];
            } else {
                sum = (((sum & 1) << 7) | ((sum & 0xfe) >> 1)) + hdr.ext[i - 8];
            }
        }
    }



   

    for (i = 0; i < lfn_blocks; i++) {



        fatMakeLfnBlock(file_name, lfn_block, lfn_blocks - i - 1, sum);


        if (i == 0)lfn_block[0] |= 0x40;
        memcopy(lfn_block, &fat_cache->data_sector[rec.hdr_idx * 32], 32);
        rec.hdr_idx++;
        if (rec.hdr_idx == 16) {
            rec.hdr_sector++;
            rec.hdr_idx = 0;
            resp = fatCacheSaveData();
            if (resp)return resp;
            resp = fatCacheLoadData(rec.hdr_sector);
            if (resp)return resp;
        }
    }



   
    /*
        if (lfn_blocks != 0) {
            resp = diskWriteSector(cache->buff_sector, cache->buff);
            if (resp)return resp;
        }
     */

    if (is_dir) {

        //ocuppy function may destroy cache buff data
        if (lfn_blocks != 0) {
            resp = diskWriteSector(fat_cache->data_sec_idx, fat_cache->data_sector);
            if (resp)return resp;
        }
        cluster = 0;
        resp = fatOccupyClusters(&cluster, 1);
        if (resp)return resp;

        fatSetClustVal(&hdr, cluster);

        fat_cache->data_sec_idx = ~0;
        resp = fatCacheLoadData(rec.hdr_sector);
        if (resp)return resp;
    }

    hdr.date = 0x5d11;
    hdr.time = 0;

    memcopy(&hdr, &fat_cache->data_sector[rec.hdr_idx * 32], 32);

    //resp = diskWriteSector(current_cache->data_sec_idx, current_cache->data_sector);
    resp = fatCacheSaveData();
    if (resp)return resp;



    if (is_dir) {

        u32 sector = fatClusterToSector(cluster);
        resp = fatClearClusters(cluster, 1);
        if (resp)return resp;

        resp = fatCacheLoadData(sector);
        if (resp)return resp;

        memfill(&hdr, 0, 32);
        //memcopy((void *)rawData, &hdr, 32);
        hdr.attrib = 0x10;
        ptr = (u8 *) & hdr;
        for (i = 0; i < 11; i++)ptr[i] = 0x20;
        ptr[0] = 0x2e;

        fatSetClustVal(&hdr, cluster);

        memcopy(&hdr, fat_cache->data_sector, 32);

        fatSetClustVal(&hdr, sub_dir_cluster);
        ptr[1] = 0x2e;

        memcopy(&hdr, fat_cache->data_sector + 32, 32);

        fatCacheSaveData();
        //fatCacheLoadData()
    }

    return 0;
}

void fatMakeLfnBlock(u8 *name_src, u8 *lfn_block, u8 block_idx, u8 crc) {

    u8 i;
    u8 name_buff[13];
    u8 *name = name_buff;
    u8 fill = 0;

    name_src += block_idx * 13;
    memcopy(name_src, name, 13);

    //fixed
    for (i = 0; i < 13; i++) {
        if (fill)name[i] = 0xff;
        if (name[i] == 0)fill = 1;
        
    }
    /*
    i++;
    for (; i < 13; i++) {
        name[i] = 0xff;
    }*/

    *lfn_block++ = block_idx + 1;

    for (i = 0; i < 5; i++) {
        *lfn_block++ = *name == 0xff ? 0xff : *name;
        *lfn_block++ = *name == 0xff ? 0xff : 0;
        name++;
    }
    *lfn_block++ = 0x0f;
    *lfn_block++ = 0;
    *lfn_block++ = crc;

    for (i = 0; i < 6; i++) {
        *lfn_block++ = *name == 0xff ? 0xff : *name;
        *lfn_block++ = *name == 0xff ? 0xff : 0;
        name++;
    }

    *lfn_block++ = 0;
    *lfn_block++ = 0;

    *lfn_block++ = *name == 0xff ? 0xff : *name;
    *lfn_block++ = *name == 0xff ? 0xff : 0;
    name++;
    *lfn_block++ = *name == 0xff ? 0xff : *name++;
    *lfn_block++ = *name == 0xff ? 0xff : 0;
    name++;

}

u8 fatGetFreeRecord(FatFullRecord *dir_rec, u8 len) {

    u8 len_ctr = 0;
    u32 sector;
    u8 resp;
    u32 current_cluster = dir_rec->entry_cluster;
    u32 next_cluster;
    u16 in_cluster_ptr = 0;
    u16 i;
    u8 is_root = 0;
    fat_cache->data_sec_idx = ~0;

    if (current_cluster == 0 && current_fat->type == FAT_TYPE_32)current_cluster = 2;

    if (current_cluster == 0) {
        //current_cluster = fatSectorToCluster(fat->root_entry);
        sector = current_fat->root_entry;
        is_root = 1;
    } else {
        sector = fatClusterToSector(current_cluster);
    }

    //VDP_drawText(APLAN, "get rec", 0, 15, cy++);

    while (len_ctr != len) {



        resp = fatCacheLoadData(sector);
        if (resp)return resp;

        for (i = 0; i < 512; i += 32) {

            if (fat_cache->data_sector[i] == 0 || fat_cache->data_sector[i] == 0xe5) {
                if (len_ctr == 0) {
                    dir_rec->hdr_sector = sector;
                    dir_rec->hdr_idx = i / 32;
                }
                len_ctr++;
            } else {
                len_ctr = 0;
                dir_rec->hdr_sector = 0;
            }
            if (len_ctr == len)return 0;
        }

        sector++;
        in_cluster_ptr++;

        if (is_root && current_fat->type == FAT_TYPE_16 && in_cluster_ptr >= 16384 / 512) {
            //VDP_drawText(APLAN, "root overflow", 0, 15, cy++);
            return FAT_ERR_ROT_OVERFLOW;
        }

        if (current_fat->type == FAT_TYPE_16 && is_root)continue;

        if (in_cluster_ptr == current_fat->cluster_size) {

            //VDP_drawText(APLAN, "cluster end", 0, 15, cy++);
            next_cluster = current_cluster;
            resp = fatGetNextCluster(&next_cluster);
            if (resp)return resp;
            if (next_cluster == 0) {
                //VDP_drawText(APLAN, "need to ocupy cluster for dir", 0, 15, cy++);
                resp = fatOccupyClusters(&current_cluster, 1);
                if (resp)return resp;
                resp = fatClearClusters(current_cluster, 1);
                if (resp)return resp;
                //VDP_drawText(APLAN, "ocupied!!!!!", TILE_ATTR(cy % 3, 0, 0, 0), 15, cy++);
                //cy %= 27;
            } else {
                current_cluster = next_cluster;
            }
            sector = fatClusterToSector(current_cluster);
            in_cluster_ptr = 0;
        }
    }

    return 0;
}

u8 fatClearClusters(u32 cluster, u8 len) {

    u32 i;
    u8 resp;
    memfill(fat_cache->data_sector, 0x00, 512);

    while (len--) {
        fat_cache->data_sec_idx = fatClusterToSector(cluster++);
        for (i = 0; i < current_fat->cluster_size; i++) {
            resp = diskWriteSector(fat_cache->data_sec_idx++, fat_cache->data_sector);
            if (resp)return resp;
        }
    }
    return 0;
}

u8 fatOccupyClusters(u32 *base_cluster, u16 len) {


    if ((current_fat->type == FAT_TYPE_16 && *base_cluster == 0xffff) || (current_fat->type == FAT_TYPE_32 && *base_cluster == 0x0FFFFFFF))return FAT_ERR_BAD_BASE_CLUSTER;
    u32 *ptr32;
    u16 *ptr16;
    u16 i;
    u32 fat_table_sector = 0;
    u16 len_ctr = 0;
    u32 first_free_cluster = 0;
    u32 cluster_ctr = 0;
    u16 fat_page_len = current_fat->type == FAT_TYPE_16 ? 256 : 128;
    u8 resp = 0;




    while (len_ctr < len) {

        resp = fatCacheLoadTable(fat_table_sector);
        if (resp)return resp;


        if (current_fat->type == FAT_TYPE_16) {

            ptr16 = (u16 *) fat_cache->table_sector;

            for (i = 0; i < fat_page_len; i++) {

                if (*ptr16++ == 0) {
                    if (first_free_cluster == 0)first_free_cluster = cluster_ctr;
                    len_ctr++;
                    if (len_ctr == len)break;
                } else if (len_ctr != 0) {
                    len_ctr = 0;
                    first_free_cluster = 0;
                }
                cluster_ctr++;

            }

            if (cluster_ctr == 0x10000)return FAT_ERR_NO_FRE_SPACE;

        } else {

            ptr32 = (u32 *) fat_cache->table_sector;
            for (i = 0; i < fat_page_len; i++) {


                if (*ptr32++ == 0) {
                    if (first_free_cluster == 0)first_free_cluster = cluster_ctr;
                    len_ctr++;
                    if (len_ctr == len)break;
                } else if (len_ctr != 0) {
                    len_ctr = 0;
                    first_free_cluster = 0;
                }
                cluster_ctr++;

            }
        }



        fat_table_sector++;
    }

    /*
    gConsDrawString("oc: ");
    gAppendNum(first_free_cluster);
     */

    if (*base_cluster != 0) {
        resp = fatSetTableRecord(*base_cluster, first_free_cluster);
        if (resp)return resp;
    }

    *base_cluster = first_free_cluster;

    while (len--) {

        if (len == 0) {
            resp = fatSetTableRecord(first_free_cluster, 0x0FFFFFFF);
        } else {
            resp = fatSetTableRecord(first_free_cluster, first_free_cluster + 1);
        }
        if (resp)return resp;
        first_free_cluster++;
    }

    resp = fatCacheSaveTable();
    if (resp)return resp;

    return 0;
}

u8 fatGetTableRecord(u32 cluster, u32 *val) {

    u8 resp;
    u8 *ptr;

    u32 table_sector = current_fat->type == FAT_TYPE_16 ? cluster / 256 : cluster / 128;
    resp = fatCacheLoadTable(table_sector);
    if (resp)return resp;

    if (current_fat->type == FAT_TYPE_16) {
        ptr = &fat_cache->table_sector[(cluster & 255) * 2];
        *val = (u32) ptr[0] | (u32) ptr[1] << 8;
    } else {
        ptr = &fat_cache->table_sector[(cluster & 127) * 4];
        *val = (u32) ptr[0] | (u32) ptr[1] << 8 | (u32) ptr[2] << 16 | (u32) ptr[3] << 24;
    }

    return 0;

}

u8 fatSetTableRecord(u32 cluster, u32 next_cluster) {

    u8 resp;
    u32 table_sector = current_fat->type == FAT_TYPE_16 ? cluster / 256 : cluster / 128;
    u8 *ptr;

    resp = fatCacheLoadTable(table_sector);
    if (resp)return resp;

    fat_cache->table_changed = 1;

    if (current_fat->type == FAT_TYPE_16) {
        ptr = &fat_cache->table_sector[(cluster & 255) * 2];
        ptr[0] = (u32) next_cluster >> 0;
        ptr[1] = (u32) next_cluster >> 8;
    } else {
        ptr = &fat_cache->table_sector[(cluster & 127) * 4];
        ptr[0] = (u32) next_cluster >> 0;
        ptr[1] = (u32) next_cluster >> 8;
        ptr[2] = (u32) next_cluster >> 16;
        ptr[3] = (u32) next_cluster >> 24;
    }

    return 0;
}

u8 fatWriteFile(void *src, u32 sectors) {

    u8 resp;
    u8 *ptr8 = (u8 *) src;
    u32 len;

    if (current_file->mode != FILE_MODE_WR)return FAT_ERR_FILE_MODE;
    if (sectors > current_file->sec_available)return FAT_ERR_OUT_OF_FILE;
    /*
    while (sectors--) {

        resp = diskWriteSector(file.sector, ptr8);
        if (resp)return resp;
        ptr8 += 512;
        resp = fatSkipSectors(1);
        if (resp)return resp;
    }*/


    while (sectors) {

        if (sectors < current_fat->cluster_size) {
            len = sectors;
        } else if (current_file->in_cluster_ptr != 0) {
            len = current_fat->cluster_size - current_file->in_cluster_ptr;
            //gConsDrawString("dbg: weird len");
        } else {
            len = current_fat->cluster_size;
        }

        resp = diskWrite(current_file->sector, ptr8, len);
        ptr8 += len * 512;
        if (resp)return resp;
        resp = fatSkipSectors(len);
        if (resp)return resp;
        sectors -= len;
    }

    if (current_file->sec_available == 0)current_file->mode = 0;
    return 0;
}

u8 fatReadFile(void *dst, u32 sectors) {

    u8 resp;
    //u32 readed;
    u32 len;

    if (current_file->mode != FILE_MODE_RD)return FAT_ERR_FILE_MODE;
    if (sectors > current_file->sec_available)return FAT_ERR_OUT_OF_FILE;

    while (sectors) {

        if (sectors < current_fat->cluster_size || current_file->in_cluster_ptr != 0) {

            len = current_fat->cluster_size - current_file->in_cluster_ptr;
            if (len > sectors)len = sectors;

            resp = diskRead(current_file->sector, dst, len);

            if (resp)return resp;
            dst += 512 * len;
            resp = fatSkipSectors(len);
            if (resp)return resp;
            sectors = sectors >= len ? sectors - len : 0;
        } else {

            resp = fatReadCluster(dst);
            if (resp)return resp;
            dst += current_fat->cluster_byte_size;
            sectors -= current_fat->cluster_size;
        }
    }

    if (current_file->sec_available == 0)current_file->mode = 0;
    return 0;
}

u8 fatReadCluster(void *dst) {

    u8 resp;
    if (current_file->mode != FILE_MODE_RD)return FAT_ERR_FILE_MODE;

    if (current_file->in_cluster_ptr != 0) {

        return fatReadFile(dst, current_fat->cluster_size - current_file->in_cluster_ptr);

    } else {

        resp = diskRead(current_file->sector, dst, current_fat->cluster_size);
        if (resp)return resp;
        dst += current_fat->cluster_byte_size;
        resp = fatGetNextCluster(&current_file->cluster);
        if (resp)return resp;
        current_file->sec_available = current_file->sec_available < current_fat->cluster_size ? 0 : current_file->sec_available - current_fat->cluster_size;
        if (current_file->sec_available != 0 && current_file->cluster == 0)return FAT_ERR_OUT_OF_TABLE;
        current_file->sector = fatClusterToSector(current_file->cluster);
        current_file->in_cluster_ptr = 0;
    }

    if (current_file->sec_available == 0)current_file->mode = 0;

    return 0;
}

u8 fatSkipSectors(u32 sectors) {

    u8 resp;

    if (sectors > current_file->sec_available)sectors = current_file->sec_available;
    while (sectors--) {

        current_file->sector++;
        current_file->sec_available--;
        current_file->in_cluster_ptr++;
        if (current_file->in_cluster_ptr == current_fat->cluster_size) {
            resp = fatGetNextCluster(&current_file->cluster);
            if (resp)return resp;
            if (current_file->cluster == 0 && sectors != 0)return FAT_ERR_OUT_OF_TABLE;
            current_file->sector = fatClusterToSector(current_file->cluster);
            current_file->in_cluster_ptr = 0;
        }
    }


    return 0;
}

u8 fatOpenFile(FatFullRecord *rec, u32 wr_sectors) {


    u8 resp;
    current_file->sec_available = rec->size / 512;
    if ((rec->size & 511) != 0)current_file->sec_available++;
    current_file->cluster = rec->entry_cluster;
    current_file->in_cluster_ptr = 0;
    current_file->mode = wr_sectors == 0 ? FILE_MODE_RD : FILE_MODE_WR;



    if (wr_sectors > current_file->sec_available) {

        resp = fatResizeFile(rec, wr_sectors);
        if (resp)return resp;
        if (current_file->cluster == 0)current_file->cluster = rec->entry_cluster;
        current_file->sec_available = wr_sectors;
    }

    current_file->sector = fatClusterToSector(current_file->cluster);


    return 0;

}

u16 fatGetRecordName(FatRecord *rec, u8 *name, u16 max_len) {

    u16 resp;
    u16 i;
    u32 saddr;

    if (max_len == 0)max_len = 256;
    u16 in_sector_addr;
    u8 *ptr;
    u8 lfn_available;


    saddr = rec->hdr_sector;
    if (rec->hdr_idx == 0)saddr--;
    in_sector_addr = rec->hdr_idx == 0 ? 480 : (rec->hdr_idx - 1) * 32;

    resp = fatCacheLoadData(saddr);
    if (resp)return resp;

    ptr = (u8 *) & fat_cache->data_sector[in_sector_addr];

    lfn_available = (*ptr & 0x0f) == 1 && ptr[0x0b] == 0x0f ? 1 : 0;

    if (!lfn_available) {

        resp = fatCacheLoadData(rec->hdr_sector);
        if (resp)return resp;
        ptr = &fat_cache->data_sector[rec->hdr_idx * 32];

        for (i = 0; i < 8; i++)name[i] = *ptr++;
        for (i = 7; name[i] == ' '; i--) {
            name[i] = 0;
            if (i == 0)break;
        }
        i++;

        name[i] = 0;

        if (ptr[0] != ' ' || ptr[1] != ' ' || ptr[2] != ' ') {
            name[i++] = '.';
            name[i++] = *ptr++;
            name[i++] = *ptr++;
            name[i++] = *ptr++;
            name[i] = 0;
        }

        return 0;
    }




    for (i = 0; i < 15; i++) {

        if (in_sector_addr == 480 || i == 0) {
            resp = fatCacheLoadData(saddr);
            if (resp)return resp;
        }


        // if ((current_cache->data_sector[in_sector_addr] & 15) > lfn_len)return FAT_LFN_BUFF_OVERFLOW;

        fatLoadLfnPart(&fat_cache->data_sector[in_sector_addr], name);
        if (fat_cache->data_sector[in_sector_addr] > 15)break;

        if (in_sector_addr == 0) {
            in_sector_addr = 480;
            saddr--;
        } else {
            in_sector_addr -= 32;
        }

    }



    ptr = name;
    while (*ptr != 0) {
        if (*ptr == 0xff)*ptr = 0;
        ptr++;
    }

    return 0;
}

u8 fatLoadLfnPart(u8 *buff_ptr, u8 *lfn_buff) {

    u16 i;

    u16 offset = *buff_ptr & 0x0f;
    if (offset == 0)return 0;
    offset = (offset - 1) * 13;
    //if(offset + 13 > )offset = 0;
    u8 *lfn_ptr = &lfn_buff[offset];
    //if (*buff_ptr == 0)return;

    //if (offset + 13 >= 256)return FAT_LFN_BUFF_OVERFLOW;


    buff_ptr++;
    for (i = 0; i < 5; i++) {
        *lfn_ptr++ = *buff_ptr++;
        buff_ptr++;
    }

    buff_ptr += 3;


    for (i = 0; i < 6; i++) {
        *lfn_ptr++ = *buff_ptr++;
        buff_ptr++;
    }

    buff_ptr += 2;

    *lfn_ptr++ = *buff_ptr++;
    buff_ptr++;
    *lfn_ptr++ = *buff_ptr++;
    //*lfn_ptr++ = 'Z';
    *lfn_ptr++ = 0;
    return 0;
}

void fatGetRecExt(FatFullRecord *rec, u8 *ext) {

    u16 len;
    u16 dot_idx;

    *ext = 0;

    len = fatSlen(rec->name);
    dot_idx = len - 1;


    while (rec->name[dot_idx] != '.') {
        if (dot_idx == 0)return;
        dot_idx--;
    }

    len = len - dot_idx - 1;
    if (len > 7 || len == 0)return;
    dot_idx++;
    while (len--) {
        *ext++ = rec->name[dot_idx++];
    }
    *ext = 0;

}

u8 fatResizeFile(FatFullRecord *rec, u32 req_sec_size) {

    u8 resp;
    u32 sec_available;
    u32 wr_clusters_req;
    u32 clusters_available;
    u32 clusters_need;
    u32 clust;
    u16 i;
    FatRecordHdr *hdr;
    u32 size;



    sec_available = rec->size / 512;
    if ((rec->size & 511) != 0)sec_available += 512;

    if (req_sec_size <= sec_available)return FAT_ERR_RESIZE;


    clusters_available = sec_available / current_fat->cluster_size;
    if ((sec_available % current_fat->cluster_size) != 0)clusters_available++;
    wr_clusters_req = req_sec_size / current_fat->cluster_size;
    if (req_sec_size % current_fat->cluster_size != 0)wr_clusters_req++;

    if (clusters_available < wr_clusters_req) {

        clusters_need = (wr_clusters_req - clusters_available);
        clust = rec->entry_cluster;
        if (clusters_available != 0) {
            for (i = 0; i < clusters_available - 1; i++) {
                fatGetNextCluster(&clust);
            }
        }

        resp = fatOccupyClusters(&clust, clusters_need);

        if (rec->entry_cluster == 0)rec->entry_cluster = clust;
        if (resp)return resp;
    }

    resp = fatCacheLoadData(rec->hdr_sector);
    if (resp)return resp;
    hdr = (FatRecordHdr *) & fat_cache->data_sector[rec->hdr_idx * 32];

    size = req_sec_size * 512;
    ((u8 *) & hdr->size)[0] = (u32) size;
    ((u8 *) & hdr->size)[1] = (u32) size >> 8;
    ((u8 *) & hdr->size)[2] = (u32) size >> 16;
    ((u8 *) & hdr->size)[3] = (u32) size >> 24;


    ((u8 *) & hdr->cluster_lo)[0] = (u32) rec->entry_cluster & 0xff;
    ((u8 *) & hdr->cluster_lo)[1] = (u32) rec->entry_cluster >> 8;
    if (current_fat->type == FAT_TYPE_32) {
        ((u8 *) & hdr->cluster_hi)[0] = (u32) rec->entry_cluster >> 16;
        ((u8 *) & hdr->cluster_hi)[1] = (u32) rec->entry_cluster >> 24;
    }

    resp = fatCacheSaveData();
    if (resp)return resp;


    return 0;
}

u8 fatGetDosRecord(void *data, u32 sector, u16 idx) {

    void *ptr;
    u8 resp;

    resp = fatCacheLoadData(sector);
    if (resp)return resp;
    ptr = &fat_cache->data_sector[idx * 32];

    memcopy(ptr, data, 32);

    return 0;
}

u8 fatGetFullRecord(FatRecord *rec, FatFullRecord *frec) {

    u8 resp;
    u8 *ptr;
    u16 idx = rec->hdr_idx;

    frec->is_dir = rec->is_dir;
    frec->hdr_idx = rec->hdr_idx;
    frec->hdr_sector = rec->hdr_sector;

    resp = fatCacheLoadData(rec->hdr_sector);
    if (resp)return resp;
    ptr = &fat_cache->data_sector[idx * 32];

    frec->entry_cluster = (u32) ptr[0x1a] | (u32) ptr[0x1b] << 8 | (u32) ptr[0x14] << 16 | (u32) ptr[0x15] << 24;
    frec->size = (u32) ptr[0x1c] | (u32) ptr[0x1d] << 8 | (u32) ptr[0x1e] << 16 | (u32) ptr[0x1f] << 24;

    resp = fatGetRecordName(rec, frec->name, 256);
    if (resp)return resp;

    return 0;
}

void fatSetCurrentFile(File *f) {
    current_file = f;
}

File* fatGetCurrentFile() {
    return current_file;
}
