/*
 * File:   fat.h
 * Author: krik
 *
 * Created on 22  2011 ., 1:06
 */

#ifndef _FAT_H
#define	_FAT_H

#include "types.h"
#include "errors.h"


#define FAT_MAX_DIR_SIZE 2048


#define FAT_TYPE_16 1
#define FAT_TYPE_32 2

typedef struct {
    u32 entry_cluster;
    u32 size;
    u32 hdr_sector;
    u16 hdr_idx;
    u8 name[256];
    u16 is_dir;
} FatFullRecord;

typedef struct {
    u32 hdr_sector;
    u8 name_begin[4];
    u16 hdr_idx;
    u16 is_dir;
} FatRecord;

typedef struct {
    u8 name[8];
    u8 ext[3];
    u8 attrib;
    u8 x[8];
    u16 cluster_hi;
    u16 time;
    u16 date;
    u16 cluster_lo;
    u32 size;
} FatRecordHdr;

typedef struct {
    u16 size;
    u32 entry_cluster;
    u32 entry_sector;
    u32 in_cluster_ptr;
    u32 current_cluster;
    FatRecord rec[FAT_MAX_DIR_SIZE];
    FatRecord * s_records[FAT_MAX_DIR_SIZE];
    u16 is_root;
} Dir;

typedef struct {
    u32 pbr_entry;
    u32 root_entry;
    u32 fat_entry;
    u32 data_entry;
    u32 loaded_sector;
    u32 cluster_byte_size;
    u32 sectors_per_fat;
    u16 type;
    u16 cluster_size;
    u16 fat_ok;
} Fat;

typedef struct {
    u8 data_sector[512];
    u8 table_sector[512];
    u32 data_sec_idx;
    u32 table_sec_idx;
    u8 table_changed;
} FatCache;

typedef struct {
    u32 sector;
    u32 cluster;
    u32 sec_available;
    u16 in_cluster_ptr;
    u8 mode;
} File;

extern Dir *dir;
extern File *current_file;

//extern FatRecord rec_tmp;

//#define ED_SAVE_DMP "/EDMD/sav-dmp.bin"
#define FILE_MODE_RD 1
#define FILE_MODE_WR 2


void fatMalloc();
u8 fatInit();
u8 fatLoadDir(u32 cluster);

u8 fatWriteFile(void *src, u32 sectors);
u8 fatReadFile(void *dst, u32 sectors);
u8 fatOpenFileByeName(u8 *name, u32 wr_sectors);
u8 fatOpenFile(FatFullRecord *rec, u32 wr_sectors);
u8 fatSkipSectors(u32 sectors);
u8 fatCreateRecIfNotExist(u8 *name, u8 is_dir);
void fatGetRecExt(FatFullRecord *rec, u8 *ext);
u8 fatGetFullRecord(FatRecord *rec, FatFullRecord *frec);
u8 fatReadDir(FatRecord *rec);
void fatOpenDir(u32 cluster);
void fatSetCurrentFile(File *f);

#endif	/* _FAT_H */

