
#include "types.h"
#include "sega.h"
#include "mem.h"
#include "graphics.h"


#define G_MAX_IMG 64

Image g_img[G_MAX_IMG];

u16 plan;
u16 cur_pal;
u16 plan_a[64 * 29];
u16 plan_b[64 * 29];

u16 plan_a_back[64 * 29];
u16 plan_b_back[64 * 29];
u16 *plan_ptr;
u16 g_vram_ptr;
u16 g_img_ctr;
u8 cy;
u16 lock_repaint;
u16 gfx_updated;

void gSelectPlan(u16 p) {
    plan = p;
}

void gAppendHex8(u8 num) {

    u8 val;
    u16 mask = (cur_pal << 13) | (TILE_FONT_BASE + 16);

    val = num >> 4;
    if (val > 9)val += 7;
    *plan_ptr++ = mask + val;
    val = num & 0x0f;
    if (val > 9)val += 7;
    *plan_ptr++ = mask + val;
}

void gAppendHex16(u16 num) {

    u16 i;
    u8 *ptr = (u8 *) & num;
    u8 val;
    u16 mask = (cur_pal << 13) | (TILE_FONT_BASE + 16);

    for (i = 0; i < 2; i++) {
        val = *ptr >> 4;
        if (val > 9)val += 7;
        *plan_ptr++ = mask + val;
        val = *ptr++ & 0x0f;
        if (val > 9)val += 7;
        *plan_ptr++ = mask + val;
    }

}

void gAppendHex32(u32 num) {

    u16 i;
    u8 *ptr = (u8 *) & num;
    u8 val;
    u16 mask = (cur_pal << 13) | (TILE_FONT_BASE + 16);

    for (i = 0; i < 4; i++) {
        val = *ptr >> 4;
        if (val > 9)val += 7;
        *plan_ptr++ = mask + val;
        val = *ptr++ & 0x0f;
        if (val > 9)val += 7;
        *plan_ptr++ = mask + val;
    }

}

void gAppendNumFlen(u32 num, u8 min_len) {

    u16 i;
    u8 buff[11];
    u8 *str = (u8 *) & buff[10];
    u16 mask = (cur_pal << 13) | (TILE_FONT_BASE + 16);

    *str = 10;
    for (i = 0; num != 0; i++) {
        *--str = num % 10;
        num /= 10;
    }

    for (; i < min_len; i++)*--str = 0;

    while (*str != 10)*plan_ptr++ = mask + *str++;
}

void gAppendNum(u32 num) {

    u16 i;
    u8 buff[11];
    u8 *str = (u8 *) & buff[10];
    u16 mask = (cur_pal << 13) | (TILE_FONT_BASE + 16);

    *str = 10;
    if (num == 0)*--str = 0;
    for (i = 0; num != 0; i++) {
        *--str = num % 10;
        num /= 10;
    }

    while (*str != 10)*plan_ptr++ = mask + *str++;
}

void gDrawNum(u16 num, u8 x, u8 y) {

    if (x > 63 || y > 27)return;
    plan_ptr = plan == APLAN ? (u16 *) & plan_a[x + (y << 6)] : (u16 *) & plan_b[x + (y << 6)];
    gAppendNum(num);
}

void gDrawNumFlen(u16 num, u8 flen, u8 x, u8 y) {

    if (x > 63 || y > 27)return;
    plan_ptr = plan == APLAN ? (u16 *) & plan_a[x + (y << 6)] : (u16 *) & plan_b[x + (y << 6)];
    gAppendNumFlen(num, flen);
}

void gDrawHex32(u32 num, u8 x, u8 y) {

    if (x > 63 || y > 27)return;
    plan_ptr = plan == APLAN ? (u16 *) & plan_a[x + (y << 6)] : (u16 *) & plan_b[x + (y << 6)];
    gAppendHex32(num);
}

void gDrawHex16(u16 num, u8 x, u8 y) {

    if (x > 63 || y > 27)return;
    plan_ptr = plan == APLAN ? (u16 *) & plan_a[x + (y << 6)] : (u16 *) & plan_b[x + (y << 6)];
    gAppendHex16(num);
}

void gDrawHex8(u16 num, u8 x, u8 y) {

    if (x > 63 || y > 27)return;
    plan_ptr = plan == APLAN ? (u16 *) & plan_a[x + (y << 6)] : (u16 *) & plan_b[x + (y << 6)];
    gAppendHex8(num);
}

void gAppendBin8(u16 num) {

    u16 mask = (cur_pal << 13) | (TILE_FONT_BASE + 16);
    u16 i;
    for (i = 0; i < 8; i++) {
        *plan_ptr++ = +mask + (num >> 7 & 1);
        num <<= 1;
    }
    gfx_updated = 1;
}

void gDrawBin8(u16 num, u8 x, u8 y) {

    if (x > 63 || y > 27)return;
    plan_ptr = plan == APLAN ? (u16 *) & plan_a[x + (y << 6)] : (u16 *) & plan_b[x + (y << 6)];
    gAppendBin8(num);
}

void gAppendString(u8 *str) {

    u16 mask = (cur_pal << 13) | (TILE_FONT_BASE - 32);

    while (*str != 0) {
        *plan_ptr++ = mask + *str++;
    }
    gfx_updated = 1;
}

void gAppendChar(u8 chr) {

    u16 mask = (cur_pal << 13) | (TILE_FONT_BASE - 32);


    *plan_ptr++ = mask + chr;
    gfx_updated = 1;
}

void gDrawString(u8 *str, u8 x, u8 y) {

    u16 mask = (cur_pal << 13) | (TILE_FONT_BASE - 32);
    if (x > 63 || y > 27)return;

    plan_ptr = plan == APLAN ? (u16 *) & plan_a[x + (y << 6)] : (u16 *) & plan_b[x + (y << 6)];

    while (*str != 0) {
        *plan_ptr++ = mask + *str++;
    }
    gfx_updated = 1;
}

void gDrawStringMaxLen(u8 *str, u8 x, u8 y, u16 len) {

    u16 mask = (cur_pal << 13) | (TILE_FONT_BASE - 32);
    if (x > 63 || y > 27)return;

    plan_ptr = plan == APLAN ? (u16 *) & plan_a[x + (y << 6)] : (u16 *) & plan_b[x + (y << 6)];

    while (*str != 0 && len--) {
        *plan_ptr++ = mask + *str++;
    }
    gfx_updated = 1;
}

void gDrawStringCn(u8 *str, u8 y) {

    u16 x = 0;
    u16 mask = (cur_pal << 13) | (TILE_FONT_BASE - 32);
    while (str[x] != 0)x++;
    //x-=2;
    x = (40 - x) / 2;
    if (x > 63 || y > 27)return;

    plan_ptr = plan == APLAN ? (u16 *) & plan_a[x + (y << 6)] : (u16 *) & plan_b[x + (y << 6)];

    while (*str != 0) {
        *plan_ptr++ = mask + *str++;
    }
    gfx_updated = 1;
}

void gCleanPlan(u8 clr_a, u8 clr_b) {

    if (clr_a) {
        mem_fill_32(0, (u32 *) plan_a, 896);
    }

    if (clr_b) {
        mem_fill_32(0, (u32 *) plan_b, 896);
    }

    gfx_updated = 1;
}

void gRepaint() {

    // static u8 ctr;


    if (lock_repaint != 0xaa55)return;
    //VDP_waitFIFOEmpty();
    if (gfx_updated) {
        VDP_waitDMACompletion();
        VDP_DoDMA(VDP_DMA_VRAM, plan_a, APLAN, 3584);
        VDP_DoDMA(VDP_DMA_VRAM, plan_b, BPLAN, 3584);
        gfx_updated = 0;
    }

    // ctr++;

}

void gSelectPalette(u16 pal) {

    cur_pal = pal & 3;
}

u16 gGetCurrentPalette() {
    return cur_pal;
}

u16 gGetCurrentPlan() {
    return plan;
}

void gFillRect(u16 tile, u8 x, u8 y, u8 w, u8 h) {

    u16 i;
    u16 *ptr = plan == APLAN ? (u16 *) & plan_a[x + (y << 6)] : (u16 *) & plan_b[x + (y << 6)];
    tile += (cur_pal << 13);

    for (i = 0; i < h; i++) {

        mem_fill_16(tile, ptr, w);
        ptr += 64;
    }
    gfx_updated = 1;
}

void gFillRectInc(u16 tile, u8 x, u8 y, u8 w, u8 h) {

    u16 i;
    u16 u;

    u16 *ptr = plan == APLAN ? (u16 *) & plan_a[x + (y << 6)] : (u16 *) & plan_b[x + (y << 6)];
    tile += (cur_pal << 13);

    for (i = 0; i < h; i++) {

        for (u = 0; u < w; u++)ptr[u] = tile++;
        //mem_fill_16(tile++, ptr, w);
        ptr += 64;
    }
    gfx_updated = 1;
}

/*
void gResetImgBuff() {
    g_vram_ptr = 32;
    g_img_ctr = 0;
}*/

void gDrawImage(Image *img, u8 x, u8 y) {


    u16 vram_addr = img->addr;
    u16 i;
    u16 *ptr = plan == APLAN ? (u16 *) & plan_a[x + (y << 6)] : (u16 *) & plan_b[x + (y << 6)];
    vram_addr += (cur_pal << 13);

    for (i = 0; i < img->h; i++) {

        mem_fill_inc_16(vram_addr, ptr, img->w);
        ptr += 64;
        vram_addr += img->w;
    }

    gfx_updated = 1;
}

/*
Image* gLoadImage(void *tiledata, u8 w, u8 h) {

    u32 len;


    len = w * h << 5;
    if (len + g_vram_ptr >= TILE_MEM_END || g_img_ctr == G_MAX_IMG)return 0;


    VDP_DoDMA(VDP_DMA_VRAM, tiledata, g_vram_ptr, len);
    g_img[g_img_ctr].w = w;
    g_img[g_img_ctr].h = h;
    g_img[g_img_ctr].addr = g_vram_ptr >> 5;
    g_img_ctr++;
    g_vram_ptr += len;
    return &g_img[g_img_ctr - 1];
}*/

void gLoadImageTo(void *tiledata, u8 w, u8 h, u16 idx_addr) {

    u32 len;
    len = w * h << 5;
    VDP_DoDMA(VDP_DMA_VRAM, tiledata, idx_addr << 5, len);

}

void gConsInit() {
    cy = 2;
    gCleanPlan(1, 1);
    gSelectPalette(0);
    gSelectPlan(APLAN);

}

void gInit() {

    gLockRepaint(1);
    cy = 2;
    gCleanPlan(1, 1);
    gSelectPalette(0);
    gSelectPlan(APLAN);

}

void gConsDrawStrNum(u8 *str, u32 num) {

    gConsDrawString(str);
    gAppendNum(num);
}

void gConsDrawString(u8 *str) {


    if (cy > 27) {
        cy = 27;
        if (plan == APLAN) {
            memcopy(&plan_a[64], plan_a, sizeof (plan_a) - 128);
            mem_fill_32(0, (u32*) (plan_a + sizeof (plan_a) - 128), 32);
        } else {
            memcopy(&plan_b[64], plan_b, sizeof (plan_a) - 128);
            mem_fill_32(0, (u32*) (plan_b + sizeof (plan_a) - 128), 32);
        }
    }

    gDrawString(str, 2, cy++);
}

void gLockRepaint(u8 state) {

    lock_repaint = state ? 0 : 0xaa55;
}

void gBackupScreen() {
    memcopy(plan_a, plan_a_back, sizeof (plan_a));
    memcopy(plan_b, plan_b_back, sizeof (plan_b));
}

void gRepairScreen() {

    memcopy(plan_a_back, plan_a, sizeof (plan_a));
    memcopy(plan_b_back, plan_b, sizeof (plan_b));
    gfx_updated = 1;
}

u16 gGetRepaintState() {

    return lock_repaint == 0 ? 0 : 1;
}
