
#include "gui.h"
#include "graphics.h"
#include "sega.h"
#include "string.h"
#include "os.h"
#include "rom.h"
#include "disk.h"
#include "mem.h"


#define IC_WIN_LEFT (VRAM_WIN_ICONS + 1)
#define IC_WIN_RIGHT (VRAM_WIN_ICONS + 2)
#define IC_WIN_LEFT_C_BOT (VRAM_WIN_ICONS + 3)
#define IC_WIN_RIGHT_C_BOT (VRAM_WIN_ICONS + 4)
#define IC_WIN_BOT (VRAM_WIN_ICONS + 5)
#define IC_WIN_TOP (VRAM_WIN_ICONS + 6)
#define IC_WIN_HD_LEFT (VRAM_WIN_ICONS + 7)
#define IC_WIN_HD_RIGHT (VRAM_WIN_ICONS + 8)
#define IC_WIN_LEFT_C_TOP (VRAM_WIN_ICONS + 9)
#define IC_WIN_RIGHT_C_TOP (VRAM_WIN_ICONS + 10)

void guiDrawWin(u8 *title, u16 x, u16 y, u16 w, u16 h) {

    u16 r_state = gGetRepaintState();

    if (r_state)gLockRepaint(1);
    gSelectPalette(0);
    gSelectPlan(BPLAN);
    gFillRect(3, x, y - 2, w, 2);
    gFillRect(1, x, y, w, h);
    // gSelectPalette(1);
    gFillRect(IC_WIN_LEFT, x - 1, y, 1, h);
    gFillRect(IC_WIN_RIGHT, x + w, y, 1, h);
    gFillRect(IC_WIN_LEFT_C_BOT, x - 1, y + h, 1, 1);
    gFillRect(IC_WIN_RIGHT_C_BOT, x + w, y + h, 1, 1);
    gFillRect(IC_WIN_BOT, x, y + h, w, 1);
    gFillRect(IC_WIN_TOP, x, y - 3, w, 1);

    gFillRect(IC_WIN_HD_LEFT, x - 1, y - 2, 1, 2);
    gFillRect(IC_WIN_HD_RIGHT, x + w, y - 2, 1, 2);

    gFillRect(IC_WIN_LEFT_C_TOP, x - 1, y - 3, 1, 1);
    gFillRect(IC_WIN_RIGHT_C_TOP, x + w, y - 3, 1, 1);

    gSelectPlan(APLAN);
    gFillRect(0, x - 1, y - 3, w + 2, h + 4);
    //gFillRect(3, x, y - 2, w, 2);
    //gFillRect(1, x, y, w, h);
    gSelectPalette(3);
    gDrawStringCn(title, y - 2);
    if (r_state)gLockRepaint(0);
}

void guiShowMessage(u8 *title, u8 *txt) {

    u16 w = strLen(txt) + 2;
    if (w < 16)w = 16;
    u16 h = 5;
    u16 x = (40 - w) / 2;
    u16 y = (28 - h) / 2;
    u16 txt_x = (40 - strLen(txt)) / 2;

    guiDrawWin(title, x, y, w, h);
    gSelectPalette(2);
    gSelectPlan(APLAN);
    gDrawString(txt, txt_x, y + 2);
    //u16 y = 28-

}

u16 guiListForm(ListForm *form) {

    //u16 x;
    u16 y;
    u16 i;
    u16 u;
    u16 first = 1;
    u16 step = 2;
    u16 txt_y = (28 - form->items_num * step) / 2;
    u16 win_w = 0;
    u16 win_x = 0;


    for (i = 0; i < form->items_num; i++) {
        u = strLen(form->items_names[i]) ;
        if (win_w < u)win_w = u;
    }

    win_w += 2;
    if (win_w < 12)win_w = 14;
    win_x = (40 - win_w) / 2;

    guiDrawWin(form->menu_name, win_x, txt_y - 1, win_w, form->items_num * 2 + 1);
    if (osIsInGame())gRepaint();

    for (;;) {


        if (!first) {

            while (osJoyRead() != 0);
            while (osJoyRead() == 0);

            if ((joy & BUTTON_UP)) {
                form->selector = form->selector == 0 ? form->items_num - 1 : form->selector - 1;
            }
            if ((joy & BUTTON_DOWN)) {
                form->selector = (form->selector + 1) % form->items_num;
            }

            if ((joy & BUTTON_B) || (joy & BUTTON_C)) {
                return 0xffff;
            }
            if ((joy & BUTTON_A)) {
                return form->selector;
            }

        }
        first = 0;


        y = txt_y;
        for (i = 0; i < form->items_num; i++) {

            if (form->selector == i) {
                gSelectPalette(2);
            } else {
                gSelectPalette(1);
            }
            gDrawStringCn(form->items_names[i], y);
            y += step;

        }

        if (osIsInGame())gRepaint();
    }


    return 0;
}

void guiDeviceInfo() {


    // static u8 * item_names[] = {"OS version", "firm version", "megakey", "system type" "assembly date", "assembly time",};

    u16 items = 7;
    u16 step = 2;
    u16 w = 26;
    u16 h = items * step - 1;
    u16 x = (40 - w) / 2;
    u16 y = (28 - h) / 2;
    u16 text_x = x + w / 2 + 3;

    gLockRepaint(1);

    guiDrawWin("System Information", x - 1, y - 1, w + 2, h + 2);

    gSelectPalette(0);


    gDrawString("Firm. Version: ", x, y);
    gDrawNum(dev_info.firm_ver, text_x, y);
    y += step;

    gDrawString("OS Version: ", x, y);
    gDrawNum(dev_info.os_ver, text_x, y);
    y += step;

    gDrawString("MegaKey Mode: ", x, y);
    if (dev_info.mkey_is_active) {
        gDrawString("Hardware", text_x, y);
    } else {
        gDrawString("Software", text_x, y);
    }
    y += step;

    gDrawString("System Type: ", x, y);
    switch (dev_info.system_region) {
        case MKEY_JAP_NTSC:
            gDrawString("JAP NTSC", text_x, y);
            break;
        case MKEY_JAP_PAL:
            gDrawString("JAP PAL", text_x, y);
            break;
        case MKEY_US_NTSC:
            gDrawString("US NTSC", text_x, y);
            break;
        case MKEY_EU_PAL:
            gDrawString("EU PAL", text_x, y);
            break;
    }

    y += step;

    gDrawString("CD-ROM: ", x, y);
    switch (dev_info.cd_region) {
        case MKEY_JAP_NTSC:
            gDrawString("JAP NTSC", text_x, y);
            break;
        case MKEY_JAP_PAL:
            gDrawString("JAP PAL", text_x, y);
            break;
        case MKEY_US_NTSC:
            gDrawString("US NTSC", text_x, y);
            break;
        case MKEY_EU_PAL:
            gDrawString("EU PAL", text_x, y);
            break;
        default:
            gDrawString("XXXX", text_x, y);
    }
    y += step;



    gDrawString("Assembly Date: ", x, y);
    gDrawString("", text_x, y);
    gAppendNum(dev_info.asm_date & 31);
    gAppendString(".");
    gAppendNum((dev_info.asm_date >> 5) & 15);
    gAppendString(".");
    gAppendNum((dev_info.asm_date >> 9) + 1980);
    y += step;

    gDrawString("Assembly Time: ", x, y);
    gDrawString("", text_x, y);
    gAppendNum(dev_info.asm_time * 2 / 60 / 60);
    gAppendString(":");
    gAppendNum(dev_info.asm_time * 2 / 60 % 60);
    gAppendString(":");
    gAppendNum(dev_info.asm_time * 2 % 60);
    y += step;
    gLockRepaint(0);

    while (joy != 0);
    while (joy == 0);
    //while (joy != 0);
}

void guiAbout() {

    CustomInfo *cutom_info = &dev_info.custom_inf;
    u16 img_detected;
    u16 i;

    u16 *pal_tmp = (u16 *) OS_TMP_BUFF;
    u16 *pal_buf = (u16 *) (OS_TMP_BUFF + 32);
    memcopy(cutom_info->pal, pal_buf, 32);

    //gCleanPlan(1, 1);
    gConsInit();

    gSelectPalette(2);
    gConsDrawString("Developed by I. Golubovskiy");
    gConsDrawString("");
    gSelectPalette(2);
    gConsDrawString("Support:");
    gSelectPalette(0);
    gConsDrawString("biokrik@gmail.com");
    gConsDrawString("http://krikzz.com");

    gConsDrawString("");
    gSelectPalette(2);
    gConsDrawString("Control:");
    gSelectPalette(0);
    gConsDrawString("LEFT/RIGHT - switch page");
    gConsDrawString("START - reload last game");
    gConsDrawString("A - file menu or start game");
    gConsDrawString("B - back");
    gConsDrawString("C - main menu");
    gConsDrawString("");


    gSelectPalette(2);
    if (cutom_info->str1[0] != 0xff)gConsDrawString(cutom_info->str1);
    if (cutom_info->str2[1] != 0xff)gConsDrawString(cutom_info->str2);
    if (cutom_info->str3[2] != 0xff)gConsDrawString(cutom_info->str3);

    img_detected = 0;
    for (i = 0; i < 16; i++)if (cutom_info->pal[i] != 0xffff)img_detected = 1;

    if (img_detected) {

        gSelectPalette(3);
        gLockRepaint(1);
        VDP_waitDMACompletion();
        VDP_readCram(48 * 2, pal_tmp, 32);
        VDP_setPalette(3, pal_buf);
        gFillRectInc(VRAM_CUST_IMG, 0, 19, 40, 8);
        gLockRepaint(0);

    }


    while (joy != 0);
    while (joy == 0);

    if (img_detected) {
        gLockRepaint(1);
        VDP_waitDMACompletion();

        VDP_setPalette(3, pal_tmp);
        gSelectPalette(0);
        gCleanPlan(1, 1);
        gLockRepaint(0);
    }
}

void guiDrawCursor(u8 *str, u16 x, u16 y) {

    static u16 cursor;

    cursor = cursor & 1;

    if (vb_ctr > 10) {

        if (cursor) {
            gDrawString(str, x, y);
        } else {
            gDrawString(" ", x, y);
        }

        vb_ctr = 0;
        cursor++;
    }

}

u16 guiHexView(FatFullRecord *rec) {

    u16 resp;
    u32 sector;
    u32 ptr = 0;
    u8 buff[512];
    u16 *ptr16;
    u8 *ptr8;
    u32 size;
    u16 i;
    u16 x = 1;
    u16 y = 1;
    u32 old_sec = 0;
    u16 repaint = 1;
    u16 u;
    //u16 key_lock = 0;
    //u16 txt_y = y + 17;

    gCleanPlan(1, 1);
    gSelectPalette(0);

    resp = fatOpenFile(rec, 0);
    if (resp)return resp;
    sector =  current_file->sector;
    size = current_file->sec_available * 512;

    vb_ctr = 0;
    for (;;) {

        if (old_sec != sector + ptr / 512) {
            old_sec = sector + ptr / 512;
            resp = diskRead(old_sec, buff, 1);
            if (resp)return resp;
        }

        if (repaint) {
            gLockRepaint(1);
            ptr16 = (u16*) & buff[ptr % 512];
            for (i = 0; i < 16; i++) {

                gSelectPalette(2);
                gDrawHex16(ptr + i * 16, x, y + i);
                gAppendString(":");
                for (u = 0; u < 4; u++) {
                    gSelectPalette(0);
                    gAppendHex16(*ptr16++);
                    gSelectPalette(1);
                    gAppendHex16(*ptr16++);
                }



            }
            gSelectPalette(1);
            gDrawString("-------------------------------------", x, y + 16);

            ptr8 = (u8*) & buff[ptr % 512];
            for (i = 0; i < 8; i++) {

                gSelectPalette(2);
                gDrawHex16(ptr + i * 32, x, y + i + 17);
                gAppendString(":");

                gSelectPalette(0);
                for (u = 0; u < 32; u++) {

                    gAppendChar(*ptr8++);
                }

            }

            repaint = 0;
            gLockRepaint(0);
        }

        while (joy != 0 && timer < 50);
        if (joy != 0) {
            timer = 40;
        } else {
            while (joy == 0);
            timer = 0;
        }

        if ((joy & BUTTON_B))return 0;

        if ((joy & BUTTON_UP)) {
            ptr = ptr == 0 ? size - 256 : ptr - 256;
            repaint = 1;
        }

        if ((joy & BUTTON_DOWN)) {
            ptr = ptr + 256 >= size ? 0 : ptr + 256;
            repaint = 1;
        }

    }

    return 0;
}
