
#include "mem.h"
#include "everdrive.h"
#include "sega.h"
#include "errors.h"
#include "string.h"
#include "graphics.h"




//u32 spi_dma_ram[128];


void mem_spi_to_ram(void *addr);
void mem_ram_to_spi(void *addr);
u8 mem_spi(u8 dat);
u8 mem_spi_dma(void *addr, u16 slen);
void mem_reload_firm();
extern u16 mem_spi_dma_code[];
extern u16 mem_reload_firm_code[];
extern u16 mem_is_empty(void *addr, u32 len, u32 val);
extern u16 mem_dma_code[];
u32 ct;

u32 malloc_ptr;
u32 malloc_ptr_68k;

#define MALLOC_BASE 0x80000

u16 memEq(void *src1, void *src2, u32 len) {

    u8 *p1 = (u8 *) src1;
    u8 *p2 = (u8 *) src2;

    while (len--)if (*p1++ != *p2++)return 0;
    return 1;
}

void *memAllocate(u32 size, u16 align) {



    if (malloc_ptr % align != 0) {
        malloc_ptr = malloc_ptr / align * align + align;
    }


    malloc_ptr += size;
    return (void *) (malloc_ptr - size);
}

u32 memRamUsed() {
    return malloc_ptr - MALLOC_BASE;
}

void memInitDmaCode() {
    memcopy(mem_dma_code, (void *) 0xff0000, 32);
}

void memInit() {
    malloc_ptr = MALLOC_BASE;

    ct = 0;

    memInitDmaCode();

}

void memcopy(void *src, void *dst, u32 len) {

    u8 mask;
    if (len == 0)return;
    mask = ((u32) src | (u32) dst | len) & 3;

    while (len) {


        if (mask == 0) {

            mem_cpy32(src, dst, len / 4);
            len = len < 65536 * 4 ? 0 : len - 65536 * 4;
            src += 65536 * 4;
            dst += 65536 * 4;

        } else if (mask == 2) {

            mem_cpy16(src, dst, len / 2);
            len = len < 65536 * 2 ? 0 : len - 65536 * 2;
            src += 65536 * 2;
            dst += 65536 * 2;

        } else {

            mem_cpy8(src, dst, len);
            len = len < 65536 ? 0 : len - 65536;
            src += 65536;
            dst += 65536;
        }
    }

}

void memfill(void *dst, u8 val, u32 len) {


    u32 val_32;
    u32 *ptr32;

    if ((((u32) dst) & 3) == 0 && (len & 15) == 0) {
        ptr32 = (u32 *) dst;
        val_32 = val << 24 | val << 16 | val << 8 | val;
        len /= 16;
        while (len--) {
            *ptr32++ = val_32;
            *ptr32++ = val_32;
            *ptr32++ = val_32;
            *ptr32++ = val_32;
        }
    } else {
        u8 *d = (u8 *) dst;
        while (len--)*d++ = val;
    }
}

u8 memSpiRead(void *dst, u16 slen) {


    u16 bank;
    u32 addr = (u32) dst;
    u16 to_rom = addr < 0x400000;
    u8 resp;
    //u16 vint_is_on;



    if (to_rom) {

        //vint_is_on = VDP_vinIsEnabled();
        addr = (u32) dst;
        bank = addr / 0x100000;
        addr &= 0xfffff;
        bank = (REG_MAP_PORT >> bank * 4) & 15;
        addr += bank * 0x100000;


        if ((addr & 511) != 0)return ERR_BAD_DMA_ADDRESS;
        //if (vint_is_on)VDP_vintOff();
        VDP_intOFF();
        resp = mem_spi_dma((void*) addr, slen);
        //if (vint_is_on)VDP_vintOn();
        VDP_intON();

        return resp;
    }

    /*
    REG_DMA_LEN = 1;
    if (to_rom) {
        addr >>= 9;
    }*/

    while (slen--) {

        if (mem_wait_fe() != 0xfe) {
            memSpiSSOff();
            return MEM_ERR_SPI_RD_TIMEOUT;
        }

        SPI16_ON;
        SPI_AR_ON;
        mem_spi_to_ram((u32*) addr);
        addr += 512;
        SPI_AR_OFF;
        SPI16_OFF;



        //mem_spi(0xff);
        //mem_spi(0xff);
    }



    return 0;
}

u8 memSpiWrite(void *src) {



    SPI16_ON;
    mem_ram_to_spi(src);
    SPI16_OFF;

    return 0;
}

void memSpiSSOn() {

    SPI_SS_ON;
}

void memSpiSSOff() {

    SPI_SS_OFF;
}

void memSpiBusy() {

    while (!IS_SPI_READY);

}

u8 memSpiIsBusy() {

    return !IS_SPI_READY;

}

void memSpiSetSpeed(u8 speed) {

    if (speed == SPI_SPEED_INIT) {
        SPI_HI_SPEED_OFF;
    } else {
        SPI_HI_SPEED_ON;
    }
}

u16 memIsEmpty(u32 val, void *addr, u32 len) {

    len /= 16;
    len -= 1;

    return mem_is_empty(addr, len, val);

}

/*
.equ app_dst_addr, 0xffff80     32bit os or client app address in sdram
.equ app_map_cfg, 0xffff84      16bit os memory mapping
.equ loader_src, 0xffff88       32bit ram loader address
.equ firm_src, 0xffff8c         32bit firmware address
.equ firm_len, 0xffff90         16bit firmware len / 4
.equ app_name, 0xffff94         string SD path of os or client app
 */

