

#include "os.h"
#include "gui.h"
#include "types.h"
#include "mem.h"
#include "sega.h"
#include "everdrive.h"
#include "string.h"
#include "graphics.h"
#include "disk.h"
#include "rom.h"
#include "usb.h"


#define OS_CFG_SIZE 2048

//variables whic stay after reboot

volatile u8 rom_start_flag;

extern u16 need_backup;
extern u16 force_sega_reinit;

extern u8 font[];
OsConfig *os_cfg;
u16 dac_type;

void restore_state();
void back_to_menu();

u32 vb_ctr;
u16 in_game_mode;
u16 snap_data_avilable;


volatile u32 timer;

#define SAV_TYPE_GAME 1
#define SAV_TYPE_CD 2

void osFlashWrite(u32 flash_addr, u16 *src, u32 len);
u16 osInitConfig();
void osStartRomIN2();

DeviceInfo dev_info;
extern u8 gfx[];
File os_file;

u16 pal_table[] = {

    0x000, 0x222, 0x000, 0xc88, 0xfff, 0x000, 0x000, 0x000, 0x000, 0x000, 0x000, 0x000, 0x000, 0x000, 0x000, 0x888,
    0x000, 0x000, 0x000, 0x000, 0x000, 0x000, 0x000, 0x000, 0x000, 0x000, 0x000, 0x000, 0x000, 0x000, 0x000, 0x2aa,
    0x000, 0x000, 0x000, 0x000, 0x000, 0x000, 0x000, 0x000, 0x000, 0x000, 0x000, 0x000, 0x000, 0x000, 0x000, 0xfff,
    0x000, 0x000, 0x000, 0x000, 0x000, 0x000, 0x000, 0x000, 0x000, 0x000, 0x000, 0x000, 0x000, 0x000, 0x000, 0x000
};

void osInitVdp() {

    u16 i;

    VDP_init();

    VDP_setPalette(0, &pal_table[16 * 0]);
    VDP_setPalette(1, &pal_table[16 * 1]);
    VDP_setPalette(2, &pal_table[16 * 2]);
    VDP_setPalette(3, &pal_table[16 * 3]);

    //gResetImgBuff();

    gLoadImageTo(font, 8, 16, TILE_FONT_BASE);

    for (i = 0; i < 16; i++) {
        memfill((void *) OS_TMP_BUFF, i << 4 | i, 32);
        gLoadImageTo((void *) OS_TMP_BUFF, 1, 1, i);
    }

    gLoadImageTo(gfx, 1, 11, VRAM_WIN_ICONS);

    REG_BANK_SRC = 2;
    gLoadImageTo((void *) 0x130080, 40, 8, VRAM_CUST_IMG);
    REG_BANK_SRC = 0;
}

u16 osInit() {


    u16 cd_present;
    u16 resp;
    //u16 i;
    rom_start_flag = 0;
    usbListenerOff();
    in_game_mode = 0;

    REG_DAC_CFG = 0;

    Z80_BUSREQ_ON;
    Z80_RESET_OFF;
    REG_MAP_PORT = 0x321f;
    


    cd_present = strEqL((void *) 0x400100, "SEGA", 4);

    if (cd_present) {
        cdInit();
    }

    PSG_init();

    memInit();
    gInit();

    fatMalloc();
    fatSetCurrentFile(&os_file);
    os_cfg = memAllocate(OS_CFG_SIZE, 512);



    //osTfcInit();

    osInitVdp();
    JOY_init();

    gConsInit();
    gLockRepaint(0);

    gConsDrawString("system init...");


    gConsDrawString("init SD...");
    diskSelect(DISK_SD);
    resp = fatInit();
    if (!resp) {
        gAppendString(" OK");
    } else {

        gAppendString("failed: ");
        gAppendNum(resp);
    }

    if (resp) {
        gConsDrawString("init MSD...");
        diskSelect(DISK_MSD);
        resp = fatInit();
        if (!resp) {
            gAppendString(" OK ");
        } else {

            gAppendString("failed: ");
            gAppendNum(resp);
        }
    }

    if (resp)return FAT_DISK_NOT_READY;

    gConsDrawString("init root...");
    resp = osInitRoot();
    if (resp)return resp;



    gConsDrawString("load config...");
    resp = osLoadConfig();
    if (resp)return resp;



    if (need_backup) {
        gConsDrawString("save backup...");
        resp = romSaveSram(need_backup == SAV_TYPE_CD);
        need_backup = 0;
        if (resp)return resp;
    }

    dev_info.system_region = CONSOLE_REGION;
    dev_info.firm_ver = REG_FIRM;
    dev_info.os_ver = 1;
    dev_info.mkey_is_active = 1;

    dev_info.cd_region = CD_NOT_PRESENT;
    if (cd_present) {
        dev_info.cd_region = romGetRegion((void *) 0x400000);
    }


    REG_MAPPER = MD_SUB_MKEY << 4;

    REG_MKEY_CFG = 0;
    if (dev_info.mkey_is_active && CONSOLE_REGION != 0)dev_info.mkey_is_active = 0;
    REG_MKEY_CFG = 1;
    if (dev_info.mkey_is_active && CONSOLE_REGION != 1)dev_info.mkey_is_active = 0;
    REG_MKEY_CFG = 2;
    if (dev_info.mkey_is_active && CONSOLE_REGION != 2)dev_info.mkey_is_active = 0;
    REG_MKEY_CFG = 3;
    if (dev_info.mkey_is_active && CONSOLE_REGION != 3)dev_info.mkey_is_active = 0;

    REG_MAPPER = 0;

    REG_BANK_SRC = 2;
    dev_info.asm_date = *(u16 *) 0x1001a4;
    dev_info.asm_time = *(u16 *) 0x1001a6;
    dev_info.serial[0] = *(u32 *) 0x1001a8;
    dev_info.serial[1] = *(u32 *) 0x1001ac;
    memcopy((void *) 0x130000, &dev_info.custom_inf, 128);
    REG_BANK_SRC = 0;
    dev_info.os_ver = *(u16 *) 0x00018e;

    YM2612_reset();
    dac_type = 0;

    usbListenerOn();

    while((joy & BUTTON_B));

    return 0;
}

u16 osInitRoot() {

    u16 resp;


    resp = fatCreateRecIfNotExist(OS_PATH_ROOT, 1);
    if (resp)return resp;

    resp = fatCreateRecIfNotExist(OS_PATH_SRAM, 1);
    if (resp)return resp;

    resp = fatCreateRecIfNotExist(OS_PATH_SNAP_SAVE, 1);
    if (resp)return resp;


    resp = fatOpenFileByeName(OS_PATH_CFG, 0);

    if (resp == FAT_ERR_NOT_EXIST) {


        /*
         * resp = fatCreateRecIfNotExist(OS_PATH_CFG, 0);
        if (resp)return resp;
        resp = fatOpenFileByeName(OS_PATH_CFG, OS_CFG_SIZE / 512);
        if (resp)return resp;
        resp = fatWriteFile(os_cfg, OS_CFG_SIZE / 512);
        if (resp)return resp;*/
        resp = osInitConfig();
        if (resp)return resp;
    }

    if (resp)return resp;

    return 0;
}

u16 osSaveConfig() {

    u16 resp;

    resp = fatOpenFileByeName(OS_PATH_CFG, OS_CFG_SIZE / 512);
    if (resp)return resp;
    resp = fatWriteFile(os_cfg, OS_CFG_SIZE / 512);
    if (resp)return resp;

    return 0;
}

u16 osLoadConfig() {

    u16 resp;

    resp = fatOpenFileByeName(OS_PATH_CFG, 0);
    if (resp)return resp;

    resp = fatReadFile(os_cfg, OS_CFG_SIZE / 512);
    if (resp)return resp;

    if (os_cfg->version != OS_CFG_VERSION) {
        osInitConfig();
    }

    return 0;
}

u16 osInitConfig() {

    u16 resp;

    memfill(os_cfg, 0, OS_CFG_SIZE);
    os_cfg->options.megakey_is_on = 1;
    os_cfg->version = OS_CFG_VERSION;
    os_cfg->options.sms_mapper = 3;

    resp = fatCreateRecIfNotExist(OS_PATH_CFG, 0);
    if (resp)return resp;
    resp = fatOpenFileByeName(OS_PATH_CFG, OS_CFG_SIZE / 512);
    if (resp)return resp;
    resp = fatWriteFile(os_cfg, OS_CFG_SIZE / 512);
    if (resp)return resp;

    return 0;
}

u16 osSelectGame(FatFullRecord *rec) {

    u16 resp;

    resp = romGenerateConfig(rec, &os_cfg->gam_cfg);
    if (resp)return resp;

    os_cfg->options.mapper = os_cfg->gam_cfg.rom_type;
    os_cfg->options.rom_region = os_cfg->gam_cfg.rom_region;

    resp = osSaveConfig();
    if (resp)return resp;

    return 0;
}

u16 osLoadRom() {


    u16 resp;
    RomConfig *cfg = &os_cfg->gam_cfg;

    gCleanPlan(1, 1);
    gSelectPalette(0);
    gDrawStringCn("Loading...", 14);
    VDP_waitVSync();
    VDP_waitVSync();
    gLockRepaint(1);
    //return 1;

    resp = fatOpenFile(&cfg->rec, 0);
    if (resp)return resp;

    //VDP_vintOff();
    VDP_intOFF();
    REG_MAP_PORT = 0x021f;
    if (current_file->sec_available * 512 > 0xb00000)return ERR_FILE_TOO_BIG;
    while (current_file->sec_available) {
        resp = fatReadFile((void *) 0x300000, current_file->sec_available < 0x100000 / 512 ? current_file->sec_available : 0x100000 / 512);
        if (resp)return resp;
        REG_MAP_PORT += 0x1000;
    }

    REG_MAP_PORT = 0x328f;

    resp = romLoadSram(os_cfg->options.mapper == MTYPE_CD);
    if (resp)return resp;



    return 0;
}


extern u16 ingame_menu_code[];
extern u8 ingame_game_vbl[];

u8 zinit_code[] = {0xF3, 0x31, 0xF0, 0xDF, 0xC3, 0x04, 0x00, 0xFF};

u16 osRunNativeCD() {

    u16 mapper = MTYPE_CD | (CD_SUB_NAT_BIOS << 4);
    u16 resp;

    gCleanPlan(1, 1);
    VDP_waitVSync();
    VDP_waitVSync();
    gLockRepaint(1);
    VDP_waitVSync();
    //VDP_vintOff();
    VDP_intOFF();


    resp = romLoadSram(1);
    if (resp)return resp;

    if (dev_info.cd_region != dev_info.system_region && dev_info.mkey_is_active && os_cfg->options.megakey_is_on) {

        mapper |= MD_SUB_MKEY << 4;
        REG_MKEY_CFG = dev_info.cd_region;
    }

    need_backup = SAV_TYPE_CD;
    in_game_mode = 1;
    run_game(mapper);

    return ERR_MUST_NOT_RET;
}


void osStartRom() {
    rom_start_flag = 1;
    VDP_setReg(1, 0x24);
    for(;;);
}

void osStartRomIN1(){
    
    if(rom_start_flag == 0)return;
    osStartRomIN2();
}


void osStartRomIN2() {


    u16 rom_type = os_cfg->options.mapper;
    u16 mapper = rom_type;
    u8 region_val;
    u16 rom_region = os_cfg->options.rom_region;
    u32 rom_size = os_cfg->gam_cfg.rec.size;
    u8 *zram;
    u16 i;
    u32 gg_addr;
    u32 gg_dat;

    
    
    gLockRepaint(1);
    while (!VDP_dmaIsComplete());
    VDP_setReg(0, 6);
    VDP_setReg(1, 4);
/*
    for (i = 2; i < 8; i++)VDP_setReg(i, 0);
    for (i = 10; i < 14; i++)VDP_setReg(i, 0);
    for (i = 15; i < 19; i++)VDP_setReg(i, 0);*/

    snap_data_avilable = 0;


    for (i = 0; i < GG_NUM; i++) {

        if (os_cfg->gam_cfg.gg_state[i] == 0)continue;

        gg_addr = os_cfg->gam_cfg.gg_addr[i];
        gg_dat = os_cfg->gam_cfg.gg_dat[i];

        if (gg_addr >= 0x400000)continue;

        REG_MAP_PORT = 0xf | (gg_addr / 0x100000 << 4);


        if (gg_addr == 0 && gg_dat == 0)continue;

        gg_addr &= 0xfffff;
        gg_addr += 0x100000;


        if (rom_type == MTYPE_SMS) {
            *((u8 *) gg_addr) = gg_dat;
        } else {

            if ((gg_addr & 1) != 0)continue;
            *((u16 *) gg_addr) = gg_dat;
        }


    }


    REG_MAP_PORT = 0x821f;


    if (rom_type == MTYPE_CD) {
        //MEM16(0x380000) = 0x0606; //set ram cart type?
        need_backup = SAV_TYPE_CD;
    }


    // sub_map = ((mapper & 0xf0) >> 4) & 0xf;
    if (rom_type == MTYPE_MD) {

        if (os_cfg->options.in_game_menu) {

            REG_MAP_PORT = 0x000f;
            MEM32(ingame_game_vbl + 2) = MEM32(0x100078);
            memcopy(ingame_menu_code, (void *) 0x100100, 128);
            memcopy(ingame_menu_code, (void *) 0x000100, 128);
            MEM32(0x100078) = 0x122;

            mapper |= MD_SUB_SNAP_SAVE_ON << 4;
            need_backup = SAV_TYPE_GAME;
        }


        if (rom_size <= 0x200200) {
            mapper |= MD_SUB_SRAM_ON << 4;
            need_backup = SAV_TYPE_GAME;
        } else if (os_cfg->gam_cfg.sram_marker == 'R') {
            mapper |= (MD_SUB_SRAM_ON | MD_SUB_SRAM_3M_ON) << 4;
            need_backup = SAV_TYPE_GAME;
        }
    }


    if (rom_type == MTYPE_32X) {

        if (rom_size <= 0x200200) {
            mapper |= MD_SUB_SRAM_ON << 4;
            need_backup = SAV_TYPE_GAME;
        } else if (os_cfg->gam_cfg.sram_marker == 'R') {
            mapper |= (MD_SUB_SRAM_ON | MD_SUB_SRAM_3M_ON) << 4;
            need_backup = SAV_TYPE_GAME;
        }

        force_sega_reinit = 1;
    }

    if (rom_type == MTYPE_SMS) {

        mapper |= (os_cfg->options.sms_mapper == 3 ? 7 : 1 << os_cfg->options.sms_mapper) << 4;
        force_sega_reinit = 1;
        Z80_BUSREQ_ON;
        Z80_RESET_OFF;
        zram = (u8 *) Z80_RAM;
        for (i = 0; i < sizeof (zinit_code); i++)*zram++ = zinit_code[i];
        Z80_RESET_ON;
        Z80_RESET_OFF;
        Z80_BUSREQ_OFF;
        for (i = 0; i < 128; i++)asm("nop");
        Z80_BUSREQ_ON;
        need_backup = SAV_TYPE_GAME;
    }



    if (rom_type != MTYPE_SMS && rom_type != MTYPE_OS && os_cfg->options.megakey_is_on != 0 && dev_info.system_region != rom_region && rom_region != MKEY_OFF) {

        if (rom_type == MTYPE_32X || dev_info.mkey_is_active == 0) {
            REG_MAP_PORT = 0x021f;
            region_val = dev_info.system_region == MKEY_JAP_NTSC ? 'J' : dev_info.system_region == MKEY_JAP_PAL ? 'J' : dev_info.system_region == MKEY_EU_PAL ? 'E' : 'U';
            MEM8(0x3001f0) = region_val;
        } else {
            mapper |= MD_SUB_MKEY << 4;
            REG_MKEY_CFG = rom_region;
        }
    }

    in_game_mode = 1;

    run_game(mapper);

}

#define JOY_DATA *((volatile u8 *) 0xa10003)
#define JOY_CONTROL *((volatile u8 *) 0xa10009)

u16 osJoyRead() {

    u8 i, j;

    if (in_game_mode) {

        VDP_waitVSync();
        JOY_DATA = 0x40;
        asm("nop");
        asm("nop");
        asm("nop");
        i = JOY_DATA & 0x3f;

        JOY_DATA = 0;

        asm("nop");
        asm("nop");
        asm("nop");
        j = (JOY_DATA & 0x30) << 2;
        joy = (~(i | j)) & 0xFF;

    }

    return joy;
}

u16 osIsInGame() {
    return in_game_mode;
}

enum {
    IG_MENU_SAVE = 0,
    IG_MENU_LOAD,
    IG_MENU_BACK_TO_MENU,
    FILE_MENU_LEN
};

void osIngameMenu() {



    u16 resp;
    ListForm form;
    SaveState *current_state = (SaveState *) SNAP_CURREN_STATE_ADDR;
    SaveState *saved_state = (SaveState *) SNAP_SAVE_ADDR;
    static u8 * item_names[] = {"save state", "load state", "back to menu"};
    form.menu_name = "ingame menu";
    form.selector = 0;
    form.items_names = item_names;
    form.items_num = FILE_MENU_LEN;

    while (!VDP_dmaIsComplete());
    osInitVdp();

    //VDP_init();
    //VDP_setReg(12, 0x81);
    gCleanPlan(1, 1);
    JOY_CONTROL = 0x40;
    gLockRepaint(0);


    resp = guiListForm(&form);

    switch (resp) {

        case IG_MENU_SAVE:
            memcopy(current_state, saved_state, 0x30000);
            snap_data_avilable = 1;
            break;
        case IG_MENU_LOAD:

            if (!snap_data_avilable) {
                snap_data_avilable = memIsEmpty(0xffffffff, (void *) SNAP_SAVE_ADDR, SAVE_LEN_SNAP) ? 0 : 1;
            }

            if (snap_data_avilable) {
                memcopy(saved_state, current_state, 0x30000);

            }
            break;
        case IG_MENU_BACK_TO_MENU:
            asm("move.w #0xabcd, (0xA13040)");
            asm("move.l 0x000004, %a0");
            asm("jmp (%a0)");
            break;
    }

    gLockRepaint(1);
    while (osJoyRead() != 0);
    VDP_waitVSync();
    restore_state();
}

/*
555 AA
2AA 55
555 80
555 AA
2AA 55
BA 30*/

u16 osInstallCustomizator(FatFullRecord *rec) {

    u16 resp = 0;
    gBackupScreen();
    u16 i;
    u16 *ptr;

    /*
        REG_BANK_SRC = 2;
        volatile u16 *flash;
        flash = (u16 *) (0x100000);
        flash[0x555] = 0xaa;
        flash[0x2aa] = 0x55;
        flash[0x555] = 0x80;
        flash[0x555] = 0xaa;
        flash[0x2aa] = 0x55;
     *(u16 *) 0x130000 = 0x30;
        // for (i = 0; i < 8192; i++) {
        while (*flash != *flash);
        // }
     */

    if (!strEq("custom.cst", rec->name))return 0;

    guiShowMessage("Customization", "Burning...");

    for (;;) {

        if (rec->size != 0x2880) {
            break;
        }

        REG_BANK_SRC = 2;
        ptr = (u16 *) 0x130000;
        for (i = 0; i < 0x2880; i += 2) {
            if (*ptr++ != 0xffff) {
                gRepairScreen();
                guiShowMessage("Customization", "Already customized!");
                while (joy != 0);
                while (joy == 0);
                gRepairScreen();
                return 0;
            }
        }
        REG_BANK_SRC = 0;

        resp = fatOpenFile(rec, 0);
        if (resp)break;
        resp = fatReadFile((void *) OS_TMP_BUFF, 21);
        if (resp)break;

        osFlashWrite(0x30000, (void *) OS_TMP_BUFF, 0x2880);

        gRepairScreen();
        guiShowMessage("Customization", "Done!!");
        while (joy != 0);
        while (joy == 0);

        break;
    }



    gRepairScreen();
    return resp;
}

//555 AA 2AA 55 555 20

void osFlashWrite(u32 flash_addr, u16 *src, u32 len) {

    REG_BANK_SRC = 2;

    volatile u16 *flash;
    u16 i;

    flash = (u16 *) (0x100000 + flash_addr);

    for (i = 0; i < len; i++)if (*flash++ != 0xffff)return;

    flash = (u16 *) 0x100000;

    flash[0x555] = 0xaa;
    flash[0x2aa] = 0x55;
    flash[0x555] = 0x20;
    flash = (u16 *) (0x100000 + flash_addr);

    for (i = 0; i < len; i++) {

        *flash = 0x00a0;
        *flash = *src++;

        while (*flash != *flash);

        flash++;
    }


    flash[0] = 0x90;
    flash[0] = 0x00;

    REG_BANK_SRC = 0;
}

void osSetMemMap(u16 map) {

    REG_MAP_PORT = map;
}

u16 osGetOsVersion() {

    return dev_info.os_ver;
}

u16 osGetFirmVersion() {

    return dev_info.firm_ver;
}

void osGetSerial(u32 *buff) {

    buff[0] = dev_info.serial[0];
    buff[1] = dev_info.serial[1];
}

u8 osReadDir(FatFullRecord *frec) {

    u8 resp;
    FatRecord rec;
    frec->entry_cluster = 0;
    resp = fatReadDir(&rec);
    if (resp)return resp;
    if (rec.hdr_sector == 0)return 0;

    resp = fatGetFullRecord(&rec, frec);
    if (resp)return resp;

    return 0;
}

u8 osSocketConnect(u8 *host, u32 port) {

    u16 i;
    u8 name_len = strLen(host);
    u8 resp = 0;
    u32 timeout = 0x20000;
    u8 buff[512];
    u8 *buff_ptr = (u8 *) buff;
    //u16 u;


    usbRead(buff, 128, 1024);
    memfill(buff, 0, 128);
    usbWrite(buff, 128, timeout);

    i = usbWrite("*cmdT", 5, timeout);
    if (i != 5)return NET_ERROR_SOPEN1;
    i = usbRead(&resp, 1, timeout);
    if (i != 1)return NET_ERROR_SOPEN2;
    if (resp != 'k')return NET_ERROR_SOPEN3;
    /*
    if (i != 1 || resp != 'k') {
        osSocketClose();


        // i = usbRead(&resp, 1, timeout);

        for (u = 16; u > 0; u--) {

            i = usbRead(&resp, 1, 65536);

            if (i != 1 || resp != '!') {
                if (u == 1)return NET_ERROR_SOPEN1;
                continue;
            }

            break;
        }

        i = usbWrite("*T", 2, timeout);
        i = usbRead(&resp, 1, timeout);
        if (i != 1 || resp != 'k')return NET_ERROR_SOPEN2;
    }*/

    *buff_ptr++ = '*';
    *buff_ptr++ = 'c';
    *buff_ptr++ = 'm';
    *buff_ptr++ = 'd';
    *buff_ptr++ = 'c';
    *buff_ptr++ = 0;
    *buff_ptr++ = 0;
    *buff_ptr++ = port >> 8;
    *buff_ptr++ = port & 0xff;
    *buff_ptr++ = name_len;
    memcopy(host, buff_ptr, name_len);
    i = usbWrite(buff, name_len + 10, timeout);
    if (i != name_len + 10)return NET_ERROR_SOPEN4;


    i = usbRead(&resp, 1, 0x800000);
    if (i == 0)return NET_ERROR_SOPEN5;
    if (resp != 'k')return NET_ERROR_SOPEN6;

    return 0;
}

u8 osSocketClose() {

    u16 i;
    u16 u;

    u8 buff[256];
    for (i = 0; i < sizeof (buff); i++)buff[i] = 0;
    i = usbWrite(buff, sizeof (buff), 0x40000);
    if (i != sizeof (buff))return NET_ERROR_CLOSE1;


    for (u = 0; u < 4; u++) {

        for (;;) {
            i = usbRead(buff, sizeof (buff), u == 0 ? 2048 : 0x8000);
            if (i != sizeof (buff))break;
        }

        i = usbWrite("*cmdT", 5, 0x10000);
        if (i != 5) continue;
        i = usbRead(buff, 1, 0x10000);
        if (i != 1)continue;
        if (buff[0] != 'k')return NET_ERROR_CLOSE2;
        return 0;
    }

    return NET_ERROR_CLOSE3;
}

u32 osSocketWrite(void *data, u32 len, u32 timeout) {

    u32 resp;
    u32 ctr = 0;
    u16 cmd;

    while (len != 0) {

        if (len < 32768) {

            cmd = 0x8000 | len;
            resp = usbWrite(&cmd, 2, timeout);
            if (resp != 2)return ctr;
            resp = usbWrite(data, len, timeout);
            ctr += resp;
            return ctr;

        } else {

            cmd = 0x8000;
            resp = usbWrite(&cmd, 2, timeout);
            if (resp != 2)return ctr;
            resp = usbWrite(data, 32768, timeout);
            ctr += resp;
            if (resp != 32768)return ctr;
            len -= resp;

        }
    }

    return ctr;
}

u32 osSocketRead(void *data, u32 len, u32 timeout) {

    return usbRead(data, len, timeout);
}

u8 osShowSnap(FatFullRecord *rec) {

    u8 resp;
    u16 i;
    SaveState *snap = (SaveState *) OS_TMP_BUFF;

    resp = fatOpenFile(rec, 0);
    if (resp)return resp;
    resp = fatReadFile((void *) OS_TMP_BUFF, SAVE_LEN_SNAP / 512);
    if (resp)return resp;

    for (i = 2; i < 17; i++) {
        if ((snap->vdp_regs[i] & 0xE000) != 0x8000)continue;
        *((volatile u16 *) GFX_CTRL_PORT) = snap->vdp_regs[i];
    }


    VDP_setReg(1, 0x04);

    VDP_DoDMA(VDP_DMA_VRAM, &snap->vdp_ram[0], 0, 32768);
    VDP_DoDMA(VDP_DMA_VRAM, &snap->vdp_ram[16384], 32768, 32768);
    VDP_DoDMA(VDP_DMA_CRAM, snap->vdp_cram, 0, 128);
    VDP_DoDMA(VDP_DMA_VSRAM, snap->vdp_vsram, 0, 80);


    VDP_setReg(1, 0x64);

    while (joy != 0);
    while (joy == 0);

    osInitVdp();

    return 0;
}

u8 osLoadSaveState(FatFullRecord *rec) {

    u8 resp;
    SaveState *snap = (SaveState *) OS_TMP_BUFF;

    resp = fatOpenFile(rec, 0);
    if (resp)return resp;
    resp = fatReadFile(snap, SAVE_LEN_SNAP / 512);
    if (resp)return resp;

    if (!strEqL(snap->key_mark, "SNAP", 4))return 0;

    memcopy(&snap->rom_cfg, &os_cfg->gam_cfg, sizeof (RomConfig));
    resp = osSaveConfig();
    if (resp)return resp;

    resp = osLoadRom();
    if (resp)return resp;

    osStartRom();

    return 0;
}
