/* 
 * File:   os.h
 * Author: krik
 *
 * Created on 4  2012 ., 4:13
 */

#ifndef _OS_H
#define	_OS_H

#include "fat.h"


#define OS_CFG_VERSION 1

#define ED_OS "/MEGA/MEGAOS.bin"
#define OS_PATH_ROOT "/MEGA"
#define OS_PATH_SRAM "/MEGA/SAVE"
#define OS_PATH_CFG "/MEGA/oscfg.bin"
#define OS_PATH_SNAP_SAVE "/MEGA/SNAP"

#define OS_TMP_BUFF 0x40000

#define GG_NUM sizeof(os_cfg->gam_cfg.gg_dat) / 2
#define CD_NOT_PRESENT 0xffff

#define SNAP_CURREN_STATE_ADDR 0x180000
#define SNAP_SAVE_ADDR 0x1b0000

#define SAVE_LEN_SNAP 0x30000
#define SAVE_LEN_32X 65536
#define SAVE_LEN_SMD 32768
#define SAVE_LEN_SMS 16384
#define SAVE_LEN_MCD 524288

#define VRAM_WIN_ICONS 16
#define VRAM_CUST_IMG 32

typedef struct {
    u16 rom_type;
    u16 rom_region;
    u16 sram_marker;
    u8 ext[4];
    u32 gg_addr[16];
    u16 gg_dat[16];
    u8 gg_state[16];
    FatFullRecord rec;
} RomConfig;

#define OPT_LEN sizeof(Options) / 2

typedef struct {
    u16 rom_region;
    u16 megakey_is_on;
    u16 in_game_menu;
    u16 mapper;
    u16 sms_mapper;

} Options;

typedef struct {
    u16 version;
    RomConfig gam_cfg;
    Options options;
} OsConfig;

typedef struct {
    u16 vdp_ram[32768];
    u16 vdp_cram[64];
    u16 vdp_regs[24];
    u16 vdp_vsram[40];
    u8 tmp[256];
} ScreenShot;

typedef struct {
    u32 cpu_regs[16];
    u16 vdp_regs[24];
    u16 cpu_ram[32768];
    u16 sram[16384];
    u16 vdp_ram[32768];
    u16 vdp_cram[64];
    u16 vdp_vsram[40];
    u8 z80_ram[8192];
    u16 z80_halt;
    u8 joy1;
    u8 joy2;
    u8 tmp[256];
    RomConfig rom_cfg;
    u8 key_mark[4];

} SaveState;

typedef struct {
    u8 str1[32];
    u8 str2[32];
    u8 str3[32];
    u16 pal[16];
} CustomInfo;

typedef struct {
    u16 mkey_is_active;
    u16 firm_ver;
    u16 os_ver;
    u16 system_region;
    u16 cd_region;
    u16 asm_date;
    u16 asm_time;
    u32 serial[2];
    CustomInfo custom_inf;
} DeviceInfo;



extern OsConfig *os_cfg;
extern DeviceInfo dev_info;
extern u32 vb_ctr;
extern volatile u32 timer;


void run_game(u8 mapper);

u16 osInit();
u16 osInitRoot();
u16 osSaveConfig();
u16 osLoadConfig();
u16 osSelectGame(FatFullRecord *rec);
u16 osLoadRom();
void osStartRom();

void osInitVdp();
void os_load_from_sdr();
void os_init_fpga();
u16 osJoyRead();
u16 osIsInGame();
u16 osRunNativeCD();
u16 osInstallCustomizator(FatFullRecord *rec);
void osSetMemMap(u16 map);
u16 osGetOsVersion();
u16 osGetFirmVersion();
void osGetSerial(u32 *buff);
u8 osReadDir(FatFullRecord *frec);
void osSetVdpReg(u8 reg, u8 val);
u8 osSocketConnect(u8 *host, u32 port);
u8 osSocketClose();
u32 osSocketWrite(void *data, u32 len, u32 timeout);
u32 osSocketRead(void *data, u32 len, u32 timeout);
u8 osShowSnap(FatFullRecord *rec);
u8 osLoadSaveState(FatFullRecord *rec);
void osStartRomIN1();
//************************************************************************OS snd
/*
u8 osPlayWav(FatFullRecord *rec);
void osTfcStartPlay(u16 play);
void osTfcUpdate(void);
void osTfcInit(const u8 *data);
 */
u8 osPlaySound(FatFullRecord *rec, u8 *ext);


#endif	/* _OS_H */

