
#include "sega.h"
#include "types.h"
#include "mem.h"
#include "everdrive.h"
#include "os.h"
#include "disk.h"
#include "graphics.h"
#include "gui.h"
#include "string.h"

#define WAV_BUFF_LEN 8


void osTfcStartPlay(u16 play);
u8 osPlayVgm(FatFullRecord *rec);

void osWavPlayerStop() {
    REG_DAC_CFG = 0;
}

u8 osWavPlayerUpdate() {

    static u16 active_buff;
    u16 resp;
    u32 addr;

    if (current_file->sec_available < WAV_BUFF_LEN / 2) {
        REG_DAC_CFG = 0;
        current_file->sec_available = 0;
        return 0;
    }

    if ((REG_DAC_CFG & 2) == 0) {

        REG_DAC_CFG = 4;
        resp = fatReadFile((void *) OS_TMP_BUFF, WAV_BUFF_LEN / 2);
        active_buff = 0x8000;

        while ((REG_DAC_CFG & 0x8000) == 0);
        while ((REG_DAC_CFG & 0x8000) != 0);

        REG_DAC_CFG = 3;
    } else {
        if (active_buff == (REG_DAC_CFG & 0x8000))return 0;
        active_buff = (REG_DAC_CFG & 0x8000);

        addr = current_file->sector;

        REG_DAC_CFG = 7;
        if (active_buff == 0x8000) {
            resp = diskRead(addr, (void *) OS_TMP_BUFF, WAV_BUFF_LEN / 2);
        } else {
            resp = diskRead(addr, (void *) OS_TMP_BUFF + 512 * WAV_BUFF_LEN / 2, WAV_BUFF_LEN / 2);
        }
        REG_DAC_CFG = 3;
        if (resp)return resp;
        resp = fatSkipSectors(WAV_BUFF_LEN / 2);
    }

    return resp;
}

u8 osPlayWav(FatFullRecord *rec) {

    u16 resp = 0;
    //REG_DAC = 0x8080;
    //REG_DAC_CFG = 0;

    current_file->sec_available = 0;

    resp = fatOpenFile(rec, 0);
    if (resp)return resp;
    resp = fatReadFile((void *) OS_TMP_BUFF, WAV_BUFF_LEN);
    if (resp)return resp;

    for (;;) {

        if (current_file->sec_available < WAV_BUFF_LEN / 2) {
            break;
        }
        resp = osWavPlayerUpdate();
        if (resp)break;
        if ((joy & (BUTTON_B | BUTTON_LEFT | BUTTON_RIGHT)))break;
    }

    osWavPlayerStop();
    return resp;


}

struct {
    u8 *data;
    u16 play;

    struct {
        u16 ptr;
        u16 wait;
        u16 freq;
        u16 retblk;
        u16 loopadr;
        u16 rep;
    } chn[6];
} TFCP;

void osTfcInit(const u8 *data) {

    Z80_BUSREQ_ON;
    Z80_RESET_OFF;

    TFCP.data = (u8*) data;
    osTfcStartPlay(0);
}

void osTfcUpdate(void) {

    u16 aa, chn, rchn, bank, key, tag, off, freq, frameptr, getptr;

    if (!TFCP.play) return;

    frameptr = 0;
    getptr = 0;

    for (chn = 0; chn < 6; chn++) {

        if (TFCP.chn[chn].wait < 0xff) {
            TFCP.chn[chn].wait++;
            continue;
        }

        if (chn < 3) {
            bank = 0;
            key = chn;
            rchn = chn;
        } else {
            bank = 1;
            key = chn + 1;
            rchn = chn - 3;
        }

        if (TFCP.chn[chn].rep > 0) {
            TFCP.chn[chn].rep--;
            if (!TFCP.chn[chn].rep) TFCP.chn[chn].ptr = TFCP.chn[chn].retblk;
        }

        while (1) {
            tag = TFCP.data[TFCP.chn[chn].ptr++];
            switch (tag) {
                case 0x7e:/*01111110 begin*/
                    TFCP.chn[chn].loopadr = TFCP.chn[chn].ptr;
                    continue;
                case 0x7f:/*01111111 end*/
                    TFCP.chn[chn].ptr = TFCP.chn[chn].loopadr;
                    TFCP.play = 0;
                    continue;
                case 0xd0:/*11010000 repeat block*/
                    TFCP.chn[chn].rep = TFCP.data[TFCP.chn[chn].ptr++];
                    off = TFCP.data[TFCP.chn[chn].ptr++] << 8;
                    off += TFCP.data[TFCP.chn[chn].ptr++];
                    TFCP.chn[chn].retblk = TFCP.chn[chn].ptr;
                    TFCP.chn[chn].ptr += (s16) off;
                    continue;
                case 0xbf:/*10111111 use old frame data disp16*/
                    off = TFCP.data[TFCP.chn[chn].ptr++] << 8;
                    off += TFCP.data[TFCP.chn[chn].ptr++];
                    frameptr = TFCP.chn[chn].ptr + (s16) off;
                    tag = TFCP.data[frameptr++];
                    break;
                case 0xff:/*11111111 use old frame data disp8*/
                    off = TFCP.data[TFCP.chn[chn].ptr++] - 256;
                    frameptr = TFCP.chn[chn].ptr + (s16) off;
                    tag = TFCP.data[frameptr++];
                    break;
                default:
                    if (tag >= 0xe0)/*skip 32..2 frames*/ {
                        TFCP.chn[chn].wait = tag;
                        break;
                    }
                    if (tag >= 0xc0)/*slide d+16*/ {
                        freq = TFCP.chn[chn].freq;
                        freq = (freq & 0xff00)+((freq + tag + 0x30)&0xff);
                        TFCP.chn[chn].freq = freq;
                        ym2612_write(0xa4 + rchn, freq >> 8, bank);
                        ym2612_write(0xa0 + rchn, freq & 0xff, bank);
                        break;
                    }
                    frameptr = TFCP.chn[chn].ptr;
                    getptr = 1;
            }

            if (frameptr) {
                if (tag & 0xc0) ym2612_write(0x28, key, 0); /*keyoff*/
                if (tag & 0x01)/*freq*/ {
                    freq = TFCP.data[frameptr++] << 8;
                    freq += TFCP.data[frameptr++];
                    TFCP.chn[chn].freq = freq;
                    ym2612_write(0xa4 + rchn, freq >> 8, bank);
                    ym2612_write(0xa0 + rchn, freq & 0xff, bank);
                }
                for (aa = 0; aa < ((tag >> 1)&0x1f); aa++)/*0..30 regs*/ {
                    ym2612_write(TFCP.data[frameptr], TFCP.data[frameptr + 1], bank);
                    frameptr += 2;
                }
                if (tag & 0x80) ym2612_write(0x28, 0xf0 | key, 0); /*keyon*/

                if (getptr) {
                    TFCP.chn[chn].ptr = frameptr;
                    getptr = 0;
                }
                frameptr = 0;
            }

            break;
        }
    }
}

void osTfcStartPlay(u16 play) {

    s16 aa, bb, cc, pp;

    TFCP.play = play;

    if (!TFCP.play) /* mute ym2612 */ {
        ym2612_write(0x2f, 0, 0); /* freq.scaler */
        ym2612_write(0x2d, 0, 0);
        ym2612_write(0x22, 0, 0); /* LFO off */
        ym2612_write(0x2b, 0, 0); /* DAC off */
        ym2612_write(0x27, 0, 0); /* CH3 normal mode */
        ym2612_write(0x27, 0, 1); /* CH3 normal mode */
        for (cc = 0; cc < 2; cc++) {
            for (aa = 0; aa < 3; aa++) {
                for (bb = 0x30; bb < 0x40; bb += 4) ym2612_write(aa + bb, 0x00, cc); /* dt1/mul */
                for (bb = 0x40; bb < 0x50; bb += 4) ym2612_write(aa + bb, 0x7f, cc); /* tl */
                for (bb = 0x50; bb < 0x60; bb += 4) ym2612_write(aa + bb, 0x00, cc); /* rs/ar */
                for (bb = 0x60; bb < 0x70; bb += 4) ym2612_write(aa + bb, 0x00, cc); /* am/d1r */
                for (bb = 0x70; bb < 0x80; bb += 4) ym2612_write(aa + bb, 0x00, cc); /* d2r */
                for (bb = 0x80; bb < 0x90; bb += 4) ym2612_write(aa + bb, 0x0f, cc); /* d1l/rr */
                for (bb = 0x90; bb < 0xa0; bb += 4) ym2612_write(aa + bb, 0x00, cc); /* ssg-eg */
                ym2612_write(0xb0 + bb, 0x00, cc); /* fb/algo */
                ym2612_write(0xb4 + bb, 0x00, cc); /* ams/fms */
                ym2612_write(0x28, aa + (cc << 2), 0); /* keyoff */
            }
        }
    } else {
        pp = 10;
        for (aa = 0; aa < 6; aa++) {
            TFCP.chn[aa].ptr = TFCP.data[pp]+(TFCP.data[pp + 1] << 8);
            TFCP.chn[aa].loopadr = TFCP.chn[aa].ptr;
            TFCP.chn[aa].wait = 0xff;
            TFCP.chn[aa].rep = 0;
            pp += 2;
        }
    }
}

u8 osPlayTfc(FatFullRecord *rec) {

    u8 resp;
    volatile u32 i;

    resp = fatOpenFile(rec, 0);
    if (resp)return resp;
    if (rec->size > 0x40000)return ERR_FILE_TOO_BIG;
    resp = fatReadFile((void *) OS_TMP_BUFF, rec->size / 512);
    if (resp)return resp;

    YM2612_reset();
    osTfcInit((u8 *) OS_TMP_BUFF);
    osTfcStartPlay(1);

    for (; TFCP.play;) {

        //VDP_waitVSync();
        for (i = 0; i < 1700; i++)asm("nop");

        if ((joy & (BUTTON_B | BUTTON_LEFT | BUTTON_RIGHT)))break;
        osTfcUpdate();
    }

    osTfcStartPlay(0);


    return 0;
}

u8 osPlaySound(FatFullRecord *frec_in, u8 *ext) {

    u8 resp = 0;
    u16 rec_idx;
    u8 prev_track = 0;

    FatFullRecord frec;


    gBackupScreen();
    guiShowMessage("Audio player", "Playback...");
    VDP_waitVSync();
    VDP_waitVSync();

    rec_idx = 0;

    for (rec_idx = 0; rec_idx < dir->size; rec_idx++) {
        if (dir->s_records[rec_idx]->hdr_idx == frec_in->hdr_idx && dir->s_records[rec_idx]->hdr_sector == frec_in->hdr_sector)break;
    }
    if (rec_idx >= dir->size)return ERR_PLAY_LIST;

    for (;;) {

        resp = fatGetFullRecord(dir->s_records[rec_idx], &frec);
        if (resp)return resp;
        fatGetRecExt(&frec, ext);


        if (strEq("wav", ext)) {
            resp = osPlayWav(&frec);
            prev_track = 0;
        } else if (strEq("tfc", ext)) {
            resp = osPlayTfc(&frec);
            prev_track = 0;
        } else if (strEq("gym", ext)) {
            resp = osPlayVgm(&frec);
            prev_track = 0;
        }

        if ((joy & BUTTON_B))break;
        if ((joy & BUTTON_LEFT)) {
            prev_track = 1;
        }
        while((joy & (BUTTON_B | BUTTON_LEFT | BUTTON_RIGHT)) != 0);
        if (prev_track) {
            rec_idx = rec_idx == 0 ? dir->size - 1 : rec_idx - 1;
        } else {
            rec_idx++;
            rec_idx %= dir->size;
        }




    }

    gRepairScreen();
    YM2612_reset();

    return resp;
}

u16 play_gym();

u8 osPlayVgm(FatFullRecord *rec) {

    u8 resp = 0;
    u8 *ptr;
    u32 len;


    //ym2612_write(0x2b, 0, 1);

    REG_MAP_PORT = 0x00ef;

    resp = fatOpenFile(rec, 0);
    if (resp)return resp;
    if (rec->size > 0xe00000)return ERR_FILE_TOO_BIG;

    while (current_file->sec_available) {
        if (current_file->sec_available < 0x100000 / 512) {
            memfill((void*) (current_file->sec_available * 512 + 0x300000), 0xff, 512);
        }
        resp = fatReadFile((void *) 0x300000, current_file->sec_available < 0x100000 / 512 ? current_file->sec_available : 0x100000 / 512);
        if (resp)return resp;
        REG_MAP_PORT += 0x1000;
    }

    // for (;;) {
    YM2612_reset();
    PSG_init();
    ym2612_write(0x2b, 0, 0);
    REG_MAP_PORT = 0x210f;

    memfill((void *) 0x100000, 0x03, 0x180);
    memfill((void *) 0x3FFFFC, 0x04, 0xff);
    len = rec->size;

    VDP_waitVSync();
    while (len > 0) {
        resp = play_gym();
        len = len > 0x300000 ? len - 0x300000 : 0;
        REG_MAP_PORT += 0x3330;
        memfill((void *) 0x3FFFFC, 0x04, 0xff);
        ptr = (u8 *) 0x100000;
        //while (*ptr > 3)*ptr++ = 4;
        if ((joy & (BUTTON_B | BUTTON_LEFT | BUTTON_RIGHT)))break;
    }
    //if ((joy & BUTTON_B))break;
    //}

    YM2612_reset();
    PSG_init();
    REG_MAP_PORT = 0x321f;
    return 0;

    return resp;
}
