

#include "rom.h"
#include "fat.h"
#include "disk.h"
#include "graphics.h"
#include "mem.h"
#include "string.h"
#include "sega.h"
#include "everdrive.h"

//{"OFF", "JAP NTSC", "JAP PAL", "US NTSC", "EU PAL"},





//RomConfig *rom_cfg;
u16 get16bitMApper(u8 *header, u8 *ext, u32 size);
u16 romSaveSramBlock(u8 *path, u8 *save_name, u16 *ptr, u16 slen);
u16 romLoadSramBlock(u8 *path, u8 *save_name, u16 *dst, u16 slen);

u16 romSaveSram(u16 cd_mode) {

    u16 slen;
    u16 resp;
    u16 map;
    u16 ram_changed = 0;
    RomConfig *rom_cfg = &os_cfg->gam_cfg;
    SaveState *snap = (SaveState *) SNAP_SAVE_ADDR;

    map = cd_mode ? MTYPE_CD : os_cfg->options.mapper;
    REG_MAP_PORT = 0x32ef;

    //slen = map == MTYPE_CD ? SAVE_LEN_MCD / 512 : map == MTYPE_SMS ? SAVE_LEN_SMS / 512 / 512 : 32768 / 512;
    switch (map) {
        case MTYPE_CD:
            slen = SAVE_LEN_MCD / 512;
            break;
        case MTYPE_SMS:
            slen = SAVE_LEN_SMS / 512;
            break;
        case MTYPE_32X:
            slen = SAVE_LEN_32X / 512;
            break;
        default:
            slen = SAVE_LEN_SMD / 512;
            break;
    }

    if (map == MTYPE_CD) {
        resp = romSaveSramBlock(OS_PATH_ROOT, "ram_cart.srm", (void *) 0x100000, slen);
    } else {
        resp = romSaveSramBlock(OS_PATH_SRAM, rom_cfg->rec.name, (void *) 0x100000, slen);
        if (resp)return resp;

        ram_changed = memIsEmpty(0xffffffff, snap, SAVE_LEN_SNAP) ? 0 : 1;

        if (os_cfg->options.in_game_menu && os_cfg->options.mapper == MTYPE_MD && ram_changed) {
            slen = SAVE_LEN_SNAP / 512;
            snap->key_mark[0] = 'S';
            snap->key_mark[1] = 'N';
            snap->key_mark[2] = 'A';
            snap->key_mark[3] = 'P';
            memcopy((void *) & os_cfg->gam_cfg, (void *) & snap->rom_cfg, sizeof (RomConfig));
            resp = romSaveSramBlock(OS_PATH_SNAP_SAVE, rom_cfg->rec.name, (void *) snap, slen);
        }
    }
    if (resp)return resp;


    return 0;
}

u16 romSaveSramBlock(u8 *path, u8 *name, u16 *src, u16 slen) {

    u16 resp;
    u16 i;
    u16 u;
    u16 ram_changed = 0;
    u8 name_buff[256];
    u32 *ptr1 = (u32 *) src;
    u32 *ptr2;

    ram_changed = memIsEmpty(0xffffffff, src, slen * 512) ? 0 : 1;
    if (!ram_changed)return 0;


    /*
    for (i = 0; i < slen; i++) {

        for (u = 0; u < 128; u++) {
            if (*ptr1++ != 0xffffffff) {
                ram_changed = 1;
                break;
            }
        }
    }*/




    i = strCopy(path, name_buff);
    name_buff[i++] = '/';
    strCopy(name, &name_buff[i]);

    // gConsDrawString(name_buff);


    resp = fatOpenFileByeName(name_buff, 0);
    if (resp == FAT_ERR_NOT_EXIST) {
        resp = fatCreateRecIfNotExist(name_buff, 0);
        if (resp)return resp;
        resp = fatOpenFileByeName(name_buff, 0);
        if (resp)return resp;
    }

    if (slen <= current_file->sec_available) {

        ram_changed = 0;
        ptr1 = (u32 *) 0x200000;
        ptr2 = (u32 *) src;
        resp = fatReadFile((void *) 0x200000, slen);
        if (resp)return resp;

        for (i = 0; i < slen; i++) {

            for (u = 0; u < 128; u++) {
                if (*ptr1++ != *ptr2++) {
                    ram_changed = 1;
                    break;
                }
            }

        }
        if (!ram_changed)return 0;
    }


    gConsDrawString("save data: ");
    gAppendHex32((u32) src);
    gAppendString(", ");
    gAppendHex32(slen * 512);

    /*
    gConsDrawString("savillllooo!");
    JOY_waitAct();*/

    resp = fatOpenFileByeName(name_buff, slen);
    if (resp)return resp;

    resp = fatWriteFile(src, slen);
    if (resp)return resp;



    return 0;
}

u16 romLoadSram(u16 cd_mode) {

    u16 resp;
    u16 map;
    u16 slen;
    RomConfig *rom_cfg = &os_cfg->gam_cfg;

    map = cd_mode ? MTYPE_CD : os_cfg->options.mapper;
    REG_MAP_PORT = 0x32ef;

    //slen = map == MTYPE_CD ? 1024 : map == MTYPE_SMS ? 16384 / 512 : 32768 / 512;
    switch (map) {
        case MTYPE_CD:
            slen = SAVE_LEN_MCD / 512;
            break;
        case MTYPE_SMS:
            slen = SAVE_LEN_SMS / 512;
            break;
        case MTYPE_32X:
            slen = SAVE_LEN_32X / 512;
            break;
        default:
            slen = SAVE_LEN_SMD / 512;
            break;
    }


    if (map == MTYPE_CD) {
        resp = romLoadSramBlock(OS_PATH_ROOT, "ram_cart.srm", (void *) 0x100000, slen);
    } else {
        resp = romLoadSramBlock(OS_PATH_SRAM, rom_cfg->rec.name, (void *) 0x100000, slen);
        if (os_cfg->options.in_game_menu && os_cfg->options.mapper == MTYPE_MD) {
            slen = SAVE_LEN_SNAP / 512;
            resp = romLoadSramBlock(OS_PATH_SNAP_SAVE, rom_cfg->rec.name, (void *) SNAP_SAVE_ADDR, slen);
            memfill((void *) SNAP_CURREN_STATE_ADDR, 0xff, SAVE_LEN_SNAP);
        }
    }
    if (resp)return resp;

    return 0;
}

u16 romLoadSramBlock(u8 *path, u8 *name, u16 *dst, u16 slen) {

    u16 resp;
    u16 i;
    u8 name_buff[256];


    i = strCopy(path, name_buff);
    name_buff[i++] = '/';
    strCopy(name, &name_buff[i]);

    resp = fatOpenFileByeName(name_buff, 0);
    if (resp == FAT_ERR_NOT_EXIST || current_file->sec_available < slen) {
        memfill(dst, 0xff, slen * 512);
        if (resp == FAT_ERR_NOT_EXIST)return 0;
    }
    if (resp)return resp;

    if (slen > current_file->sec_available)slen = current_file->sec_available;
    resp = fatReadFile(dst, slen);
    if (resp)return resp;

    return 0;
}

u16 romGenerateConfig(FatFullRecord *rec, RomConfig *rom_cfg) {

    //u16 i;
    u8 buff[512];
    u16 resp;

    //gConsInit();
    //fatGetRecordName(rec, rom_cfg->name, 0);


    //strCopy(rec->name, rom_cfg->name);
    rom_cfg->rom_region = 0;
    //rom_cfg->entry_cluster = rec->entry_cluster;
    rom_cfg->rom_type = MTYPE_MD;
    rom_cfg->ext[0] = 'u';
    rom_cfg->ext[1] = 'k';
    rom_cfg->ext[2] = 0;
    rom_cfg->ext[3] = 0;
    rom_cfg->sram_marker = 0;

    /*
    i = strLen(rom_cfg->name);
    if (i < 3)return 0;
    if (rom_cfg->name[i - 3] != '.' && rom_cfg->name[i - 4] != '.') {
        return 0;
    }
    i -= 2;
    if (rom_cfg->name[i - 1] != '.')i--;

    memcopy(&rom_cfg->name[i], rom_cfg->ext, 3);
     */
    memcopy(rec, &rom_cfg->rec, sizeof (FatFullRecord));
    fatGetRecExt(rec, rom_cfg->ext);

    //gConsDrawString(&rom_cfg->name[i]);
    //gConsDrawString(rom_cfg->ext);


    if (strEq(rom_cfg->ext, "smd") || strEq(rom_cfg->ext, "bin") || strEq(rom_cfg->ext, "gen") || strEq(rom_cfg->ext, "32x")) {

        //gConsDrawString("16 bit rom");
        resp = fatOpenFile(rec, 0);
        if (resp)return resp;
        resp = fatReadFile(buff, 1);
        if (resp)return resp;
        rom_cfg->rom_type = get16bitMApper(buff, rom_cfg->ext, rec->size);
        //gConsDrawString("size: ");
        //gAppendNum(rec->size);
        rom_cfg->rom_region = romGetRegion(buff);

        rom_cfg->sram_marker = buff[0x1b0];
        if (strEqL("SONIC & KNUCKLES", &buff[0x120], 16))rom_cfg->sram_marker = 'R';

        //if (rom_cfg->rom_region != 0xffff && rom_cfg->rom_region != CONSOLE_REGION)rom_cfg->rom_type |= MD_SUB_MKEY << 4;

    } else if (strEq(rom_cfg->ext, "sms") || strEq(rom_cfg->ext, "sg") || strEq(rom_cfg->ext, "sc") || strEq(rom_cfg->ext, "gg")) {

        //gConsDrawString("8 bit rom");
        rom_cfg->rom_type = MTYPE_SMS; // | (SMS_SUB_STD | SMS_SUB_KOR | SMS_SUB_CDM) << 4;

    } else {

        //gConsDrawString("unknown rom");
    }

    //memcopy(rec, &rom_cfg->rec, sizeof (FatRecord));
    memfill(rom_cfg->gg_addr, 0, sizeof (rom_cfg->gg_addr));
    memfill(rom_cfg->gg_dat, 0, sizeof (rom_cfg->gg_dat));
    memfill(rom_cfg->gg_state, 0, sizeof (rom_cfg->gg_state));

    return 0;
}

u16 romGetRegion(u8 *rom_hdr) {

    u8 val = rom_hdr[0x1f0];
    if (val != rom_hdr[0x1f1] && rom_hdr[0x1f1] != 0x20 && rom_hdr[0x1f1] != 0)return MKEY_OFF;

    switch (val) {

        case 'F':
            return MKEY_OFF;
        case 'U':
        case 'W':
        case '4':
        case 4:
            return MKEY_US_NTSC;
        case 'J':
        case 'B':
        case '1':
        case 1:
            return MKEY_JAP_NTSC;
        case 'E':
        case 'A':
        case '8':
        case 8:
            return MKEY_EU_PAL;

    }

    return 2;

}

u16 get16bitMApper(u8 *header, u8 *ext, u32 size) {

    u8 *hdr_name = (u8 *) & header[0x100];

    hdr_name[0x50] = 0;

    if (size > 0x400000 && strContain("SUPER STREET FIGHTER2", hdr_name))return MTYPE_SSF;
    if (size > 0x400000)return MTYPE_10M;
    if (size == 0x20000 && strContain("BOOT ROM", hdr_name))return MTYPE_CD;
    if (strContain("SEGA 32X", hdr_name) || strEq(ext, "32x")) return MTYPE_32X;
    if (strContain("SEGA EVERDRIVE", hdr_name)) return MTYPE_OS;

    return MTYPE_MD;
    /*
    if (size <= 0x200200) {
        return MTYPE_MD | (MD_SUB_SRAM_ON << 4);
    } else if (header[0x1b0] == 'R') {
        return MTYPE_MD | (MD_SUB_SRAM_ON << 4) | (MD_SUB_SRAM_3M_ON << 4);
    } else {
        return MTYPE_MD;
    }*/

}

/*
const u8 def_reg_state[] = {
    0x04, 0x04, 0x30, 0x2c, 0x07, 0x54, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x81, 0x2b, 0x00, 0x02, 0x01, 0x00, 0x00, 0xff, 0xff
};*/

