
#include "types.h"
#include "everdrive.h"
#include "usb.h"
#include "mem.h"
#include "sega.h"
#include "graphics.h"
#include "os.h"


void usbLoadOs();
void usbLoadFirmware();
u16 usbRunSelectedGame();
void usbRunSmsGame();
void usbRunSmdGame();
void usbRunGame();
u16 usbLoadGame();

u16 usb_listener_on;

u8 usbReadByte() {

    while ((STATE_PORT & (1 << 2)) == 0);
    return FIFO_PORT;
}

u16 usbReadWord() {

    volatile u16 dat;
    while ((STATE_PORT & (1 << 2)) == 0);
    dat = FIFO_PORT;
    while ((STATE_PORT & (1 << 2)) == 0);
    return FIFO_PORT;
}

u8 usbReadPio(u16 *dst, u16 slen) {

    u16 i;

    while (slen--) {

        for (i = 0; i < 256; i++) {
            while ((STATE_PORT & (1 << 2)) == 0);
            *dst = FIFO_PORT;
            while ((STATE_PORT & (1 << 2)) == 0);
            *dst++ = FIFO_PORT;
        }
    }

    return 0;
}

u8 usbReadDma(u16 *dst, u16 slen) {

    u16 bank;
    //u16 vint_is_on = VDP_vinIsEnabled();
    u32 addr; // = (u32) dst;
    //if (vint_is_on)VDP_vintOff();
    VDP_intOFF();
    //addr >>= 9;

    addr = (u32) dst;

    bank = addr / 0x100000;
    addr &= 0xfffff;
    bank = (REG_MAP_PORT >> bank * 4) & 15;
    addr += bank * 0x100000;

    if ((addr & 511) != 0)return ERR_BAD_DMA_ADDRESS;

    REG_DMA_LEN = slen;
    REG_DMA_ADDR = addr / 512;
    run_dma(DMA_USB);
    while ((STATE_PORT & 1 << 3) == 0);
    VDP_intON();

    return 0;
}

void usbWriteByte(u8 dat) {

    while ((STATE_PORT & (1 << 1)) == 0);
    FIFO_PORT = dat;
}

u32 usbWrite(void *data, u32 len, u32 timeout) {

    u8 *ptr8 = (u8 *) data;
    u32 byt_ctr;
    u32 ctr;

    if (timeout == 0) {
        byt_ctr = len;
        while (len--) {
            while ((STATE_PORT & (1 << 1)) == 0);
            FIFO_PORT = *ptr8++;
        }

    } else {
        byt_ctr = 0;

        while (len--) {

            ctr = timeout;
            while ((STATE_PORT & (1 << 1)) == 0) {
                if (ctr-- == 0)return byt_ctr;
            }
            FIFO_PORT = *ptr8++;
            byt_ctr++;
        }
    }

    return byt_ctr;
}

u32 usbRead(void *data, u32 len, u32 timeout) {

    u8 *ptr8 = (u8 *) data;
    u32 byt_ctr;
    u32 ctr;

    if (timeout == 0) {
        byt_ctr = len;
        while (len--) {
            while ((STATE_PORT & (1 << 2)) == 0);
            *ptr8++ = FIFO_PORT;
        }
    } else {

        byt_ctr = 0;

        while (len--) {

            ctr = timeout;
            while ((STATE_PORT & (1 << 2)) == 0) {
                if (ctr-- == 0)return byt_ctr;
            }
            *ptr8++ = FIFO_PORT;
            byt_ctr++;
        }

    }
    return byt_ctr;
}

u8 usbRdReady() {

    return STATE_PORT & (1 << 2);
}

u8 usbWrReady() {

    return STATE_PORT & (1 << 1);
}

void usbListener() {

    u8 dat;

    if (usb_listener_on == 0)return;
    if (!usbRdReady())return;
    if (usbReadByte() != '*')return;

    dat = usbReadByte();

    usbListenerOff();
    switch (dat) {

        case 'T':
            usbWriteByte('k');
            break;
        case 'w':
        case 'o':
            usbLoadOs();
            break;
        case 'f':
            usbLoadFirmware();
            break;

        case 'g':
            gBackupScreen();
            usbLoadGame();
            gRepairScreen();
            break;

        case 's':
            gBackupScreen();
            usbRunSelectedGame();
            gRepairScreen();
            break;
        case 'r':
            usbRunGame();
            break;
    }
    usbListenerOn();
}

void usbLoadFirmware() {


    u32 slen = usbReadByte();
    slen *= 2;
    gLockRepaint(1);
    while (!VDP_dmaIsComplete());
    //VDP_vintOff();
    VDP_intOFF();


    usbWriteByte('k');
    usbReadDma((void *) 0x008400, (u16) slen);
    usbWriteByte('d');

    os_init_fpga();
}

void usbLoadOs() {


    u16 slen = usbReadByte();
    gLockRepaint(1);
    while (!VDP_dmaIsComplete());
    //VDP_vintOff();
    VDP_intOFF();
    REG_MAP_PORT = 0x32Ef;

    usbWriteByte('k');
    slen *= 128;
    usbReadDma((void *) 0x100000, slen);
    usbWriteByte('d');

    VDP_setPaletteColor(0, 0, 0x0f0);

    os_load_from_sdr();
}

void usbListenerOff() {
    usb_listener_on = 0;
}

void usbListenerOn() {
    usb_listener_on = 1;
}

u16 usbRunSelectedGame() {

    u16 resp;


    gCleanPlan(1, 1);
    gDrawStringCn("loading...", 13);
    gRepaint();

    usbWriteByte('k');

    resp = osLoadRom();
    if (resp)return resp;
    osStartRom();

    return 0;
}

void usbRunGame() {

    u8 mapper = 0;
    u8 dat = usbReadByte();

    switch (dat) {

        case 's':
            mapper = MTYPE_SMS | (7 << 4);
            break;
        case 'm':
            mapper = MTYPE_MD;
            break;
        case 'o':
            mapper = MTYPE_OS;
            break;
        case 'c':
            mapper = MTYPE_CD;
            break;
        case 'M':
            mapper = MTYPE_10M;
            break;

        default:
            return;

    }


    usbWriteByte('k');

    gLockRepaint(1);
    while (!VDP_dmaIsComplete());
    VDP_setReg(1, 6);
    VDP_setReg(1, 4);

    run_game(mapper);
}

u16 usbLoadGame() {

    u16 slen = usbReadByte();


    gCleanPlan(1, 1);
    gDrawStringCn("loading...", 13);
    gRepaint();

    usbWriteByte('k');

    slen *= 128;
    REG_MAP_PORT = 0x210f;
    usbReadDma((void *) 0x100000, slen);

    usbWriteByte('d');

    return 0;
}
